"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

===============================

MNIST classifier implementation that demonstrate end to end training, inference and evaluation using FuseMedML
"""

import copy
import logging
import os
from typing import OrderedDict, Tuple
from fuse.dl.lightning.pl_funcs import convert_predictions_to_dataframe

import torch
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data.dataloader import DataLoader
import pytorch_lightning as pl

from fuse.eval.evaluator import EvaluatorDefault
from fuse.eval.metrics.classification.metrics_thresholding_common import MetricApplyThresholds
from fuse.eval.metrics.classification.metrics_classification_common import MetricAccuracy, MetricAUCROC, MetricROCCurve

from fuse.data.utils.samplers import BatchSamplerDefault
from fuse.data.utils.collates import CollateDefault

from fuse.dl.losses.loss_default import LossDefault
from fuse.dl.models.model_wrapper import ModelWrapSeqToDict
from fuse.dl.lightning.pl_module import LightningModuleDefault

from fuse.utils.utils_debug import FuseDebug
from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils.file_io.file_io import create_dir, save_dataframe
import fuse.utils.gpu as GPU

from fuseimg.datasets.mnist import MNIST

from fuse_examples.imaging.classification.mnist import lenet

###########################################################################################################
# Fuse
###########################################################################################################

##########################################
# Debug modes
##########################################
mode = "default"  # Options: 'default', 'debug'. See details in FuseDebug
debug = FuseDebug(mode)

##########################################
# Output Paths
##########################################
ROOT = "_examples/mnist"  # TODO: fill path here
model_dir = os.path.join(ROOT, "model_dir")
PATHS = {
    "model_dir": model_dir,
    "cache_dir": os.path.join(ROOT, "cache_dir"),
    "inference_dir": os.path.join(model_dir, "infer_dir"),
    "eval_dir": os.path.join(model_dir, "eval_dir"),
}

NUM_GPUS = 1
##########################################
# Train Common Params
##########################################
TRAIN_COMMON_PARAMS = {}

# ============
# Data
# ============
TRAIN_COMMON_PARAMS["data.batch_size"] = 100
TRAIN_COMMON_PARAMS["data.train_num_workers"] = 8
TRAIN_COMMON_PARAMS["data.validation_num_workers"] = 8

# ===============
# PL Trainer
# ===============
TRAIN_COMMON_PARAMS["trainer.num_epochs"] = 2
TRAIN_COMMON_PARAMS["trainer.num_devices"] = NUM_GPUS
TRAIN_COMMON_PARAMS["trainer.accelerator"] = "gpu"
# use "dp" strategy temp when working with multiple GPUS - workaround for pytorch lightning issue: https://github.com/Lightning-AI/lightning/issues/11807
TRAIN_COMMON_PARAMS["trainer.strategy"] = "dp" if TRAIN_COMMON_PARAMS["trainer.num_devices"] > 1 else None
TRAIN_COMMON_PARAMS["trainer.ckpt_path"] = None  # path to the checkpoint you wish continue the training from

# ===============
# Optimizer
# ===============
TRAIN_COMMON_PARAMS["opt.lr"] = 1e-4
TRAIN_COMMON_PARAMS["opt.weight_decay"] = 0.001


def perform_softmax(logits: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    cls_preds = F.softmax(logits, dim=1)
    return logits, cls_preds


def create_model() -> torch.nn.Module:
    torch_model = lenet.LeNet()
    # wrap basic torch model to automatically read inputs from batch_dict and save its outputs to batch_dict
    model = ModelWrapSeqToDict(
        model=torch_model,
        model_inputs=["data.image"],
        post_forward_processing_function=perform_softmax,
        model_outputs=["model.logits.classification", "model.output.classification"],
    )
    return model


#################################
# Train Template
#################################
def run_train(paths: dict, train_params: dict):
    # ==============================================================================
    # Logger
    # ==============================================================================
    fuse_logger_start(output_path=paths["model_dir"], console_verbose_level=logging.INFO)
    print("Fuse Train")

    # ==============================================================================
    # Data
    # ==============================================================================
    # Train Data
    print("Data - trainset:")
    train_dataset = MNIST.dataset(paths["cache_dir"], train=True)
    print("- Create sampler:")
    sampler = BatchSamplerDefault(
        dataset=train_dataset,
        balanced_class_name="data.label",
        num_balanced_classes=10,
        batch_size=train_params["data.batch_size"],
        balanced_class_weights=None,
    )
    print("- Create sampler: Done")

    # Create dataloader
    train_dataloader = DataLoader(
        dataset=train_dataset,
        batch_sampler=sampler,
        collate_fn=CollateDefault(),
        num_workers=train_params["data.train_num_workers"],
    )
    print("Data - trainset: Done")

    ## Validation data
    print("Data - validation set:")
    # wrapping torch dataset
    validation_dataset = MNIST.dataset(paths["cache_dir"], train=False)

    # dataloader
    validation_dataloader = DataLoader(
        dataset=validation_dataset,
        batch_size=train_params["data.batch_size"],
        collate_fn=CollateDefault(),
        num_workers=train_params["data.validation_num_workers"],
    )
    print("Data - validation set: Done")

    # ====================================================================================
    # Model
    # ====================================================================================
    model = create_model()

    # ====================================================================================
    # Losses
    # ====================================================================================
    losses = {
        "cls_loss": LossDefault(
            pred="model.logits.classification", target="data.label", callable=F.cross_entropy, weight=1.0
        ),
    }

    # ====================================================================================
    # Metrics
    # ====================================================================================
    train_metrics = OrderedDict(
        [
            ("operation_point", MetricApplyThresholds(pred="model.output.classification")),  # will apply argmax
            ("accuracy", MetricAccuracy(pred="results:metrics.operation_point.cls_pred", target="data.label")),
        ]
    )

    validation_metrics = copy.deepcopy(train_metrics)  # use the same metrics in validation as well

    # either a dict with arguments to pass to ModelCheckpoint or list dicts for multiple ModelCheckpoint callbacks (to monitor and save checkpoints for more then one metric).
    best_epoch_source = dict(
        monitor="validation.metrics.accuracy",
        mode="max",
    )

    # ====================================================================================
    # Training components
    # ====================================================================================
    # create optimizer
    optimizer = optim.Adam(model.parameters(), lr=train_params["opt.lr"], weight_decay=train_params["opt.weight_decay"])

    # create learning scheduler
    lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer)
    lr_sch_config = dict(scheduler=lr_scheduler, monitor="validation.losses.total_loss")

    # optimizier and lr sch - see pl.LightningModule.configure_optimizers return value for all options
    optimizers_and_lr_schs = dict(optimizer=optimizer, lr_scheduler=lr_sch_config)

    # ====================================================================================
    # Train
    # ====================================================================================
    # create instance of PL module - FuseMedML generic version
    pl_module = LightningModuleDefault(
        model_dir=paths["model_dir"],
        model=model,
        losses=losses,
        train_metrics=train_metrics,
        validation_metrics=validation_metrics,
        best_epoch_source=best_epoch_source,
        optimizers_and_lr_schs=optimizers_and_lr_schs,
    )

    # create lightining trainer.
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        max_epochs=train_params["trainer.num_epochs"],
        accelerator=train_params["trainer.accelerator"],
        strategy=train_params["trainer.strategy"],
        devices=train_params["trainer.num_devices"],
        auto_select_gpus=True,
    )

    # train
    pl_trainer.fit(pl_module, train_dataloader, validation_dataloader, ckpt_path=train_params["trainer.ckpt_path"])
    print("Train: Done")


######################################
# Inference Common Params
######################################
INFER_COMMON_PARAMS = {}
INFER_COMMON_PARAMS["infer_filename"] = "infer_file.gz"
INFER_COMMON_PARAMS["checkpoint"] = "best_epoch.ckpt"
INFER_COMMON_PARAMS["trainer.num_devices"] = 1
INFER_COMMON_PARAMS["trainer.accelerator"] = "gpu"
INFER_COMMON_PARAMS["trainer.strategy"] = None

######################################
# Inference Template
######################################


def run_infer(paths: dict, infer_common_params: dict):
    create_dir(paths["inference_dir"])
    infer_file = os.path.join(paths["inference_dir"], infer_common_params["infer_filename"])
    checkpoint_file = os.path.join(paths["model_dir"], infer_common_params["checkpoint"])
    #### Logger
    fuse_logger_start(output_path=paths["model_dir"], console_verbose_level=logging.INFO)

    print("Fuse Inference")
    print(f"infer_filename={infer_file}")

    ## Data
    # Create dataset
    validation_dataset = MNIST.dataset(paths["cache_dir"], train=False)
    # dataloader
    validation_dataloader = DataLoader(
        dataset=validation_dataset, collate_fn=CollateDefault(), batch_size=2, num_workers=2
    )

    # load pytorch lightning module
    model = create_model()
    pl_module = LightningModuleDefault.load_from_checkpoint(
        checkpoint_file, model_dir=paths["model_dir"], model=model, map_location="cpu", strict=True
    )
    # set the prediction keys to extract (the ones used be the evaluation function).
    pl_module.set_predictions_keys(
        ["model.output.classification", "data.label"]
    )  # which keys to extract and dump into file

    print("Model: Done")
    # create a trainer instance
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        accelerator=infer_common_params["trainer.accelerator"],
        devices=infer_common_params["trainer.num_devices"],
        strategy=infer_common_params["trainer.strategy"],
        auto_select_gpus=True,
    )
    predictions = pl_trainer.predict(pl_module, validation_dataloader, return_predictions=True)

    # convert list of batch outputs into a dataframe
    infer_df = convert_predictions_to_dataframe(predictions)
    save_dataframe(infer_df, infer_file)


######################################
# Analyze Common Params
######################################
EVAL_COMMON_PARAMS = {}
EVAL_COMMON_PARAMS["infer_filename"] = INFER_COMMON_PARAMS["infer_filename"]


######################################
# Eval Template
######################################
def run_eval(paths: dict, eval_common_params: dict):
    create_dir(paths["eval_dir"])
    infer_file = os.path.join(paths["inference_dir"], eval_common_params["infer_filename"])
    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)

    print("Fuse Eval")

    # metrics
    class_names = [str(i) for i in range(10)]

    metrics = OrderedDict(
        [
            ("operation_point", MetricApplyThresholds(pred="model.output.classification")),  # will apply argmax
            ("accuracy", MetricAccuracy(pred="results:metrics.operation_point.cls_pred", target="data.label")),
            (
                "roc",
                MetricROCCurve(
                    pred="model.output.classification",
                    target="data.label",
                    class_names=class_names,
                    output_filename=os.path.join(paths["eval_dir"], "roc_curve.png"),
                ),
            ),
            ("auc", MetricAUCROC(pred="model.output.classification", target="data.label", class_names=class_names)),
        ]
    )

    # create evaluator
    evaluator = EvaluatorDefault()

    # run
    results = evaluator.eval(ids=None, data=infer_file, metrics=metrics, output_dir=paths["eval_dir"])

    return results


######################################
# Run
######################################
if __name__ == "__main__":
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    GPU.choose_and_enable_multiple_gpus(NUM_GPUS, force_gpus=force_gpus)

    RUNNING_MODES = ["train", "infer", "eval"]  # Options: 'train', 'infer', 'eval'
    # train
    if "train" in RUNNING_MODES:
        run_train(paths=PATHS, train_params=TRAIN_COMMON_PARAMS)

    # infer
    if "infer" in RUNNING_MODES:
        run_infer(paths=PATHS, infer_common_params=INFER_COMMON_PARAMS)

    # eval
    if "eval" in RUNNING_MODES:
        run_eval(paths=PATHS, eval_common_params=EVAL_COMMON_PARAMS)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import inspect
import logging
import os
import sys
import threading
import time
from collections import defaultdict
from logging.handlers import RotatingFileHandler
from multiprocessing import current_process
from shutil import copyfile
from typing import Dict, Optional, Any, List

from termcolor import colored

from fuse.utils.file_io.file_io import create_dir, create_or_reset_dir


class ProcessSafeHandler(logging.StreamHandler):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._locks = defaultdict(lambda: threading.RLock())

    def acquire(self):
        current_process_id = current_process().pid
        self._locks[current_process_id].acquire()

    def release(self):
        current_process_id = current_process().pid
        self._locks[current_process_id].release()


class ConsoleFormatter(logging.Formatter):
    """Logging Formatter to add colors per verbose level and file, line number in case of warnning/error"""

    grey = "\x1b[38;21m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    warn_format = "(%(filename)s:%(lineno)d) WARNING: %(message)s"
    err_format = "(%(filename)s:%(lineno)d) ERROR: %(message)s"

    LEVEL_FORMATS = {
        logging.WARNING: yellow + warn_format + reset,
        logging.ERROR: red + err_format + reset,
        logging.CRITICAL: bold_red + err_format + reset,
    }

    def format(self, record):
        if record.levelno in [logging.INFO, logging.DEBUG]:
            if isinstance(record.__dict__["args"], dict):
                color = record.__dict__["args"].get("color", None)
                on_color = record.__dict__["args"].get("on_color", None)
                attrs = record.__dict__["args"].get("attrs", None)
                if attrs is not None and not isinstance(attrs, list):
                    attrs = [attrs]
                log_fmt = colored("%(message)s", color=color, on_color=on_color, attrs=attrs)
            else:
                log_fmt = "%(message)s"
        else:
            log_fmt = self.LEVEL_FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)

        return formatter.format(record)


def fuse_logger_start(
    output_path: Optional[str] = None,
    console_verbose_level: int = logging.INFO,
    list_of_source_files: Optional[List[str]] = None,
) -> None:
    """
    Start Fuse logger:
    Define output destination; console, file and file including also debug
    Define the formats:
    - Console - color per verbose level and (file:line) prefix in a case of warning or error
    - File - adds prefix '(file:line) : level'
    - Debug file - adds prefix 'time: (file:line) : level'

    :param output_path: place to save the log files. If not specified log to files will be disabled
    :param console_verbose_level: verbose level used to print to the screen
    :param force_reset: if logs already exist in output_path - delete automatically (True) or prompt user before deletion, default is False
    :param list_of_source_files: list of source files or paths to save (relative path to the working directory)
           If None, just the file of the caller fuction will be saved

    :return: None
    """
    lgr = logging.getLogger("Fuse")
    # skip if already configured
    if lgr.level != 0:
        return
    lgr.setLevel(logging.DEBUG)

    # console
    console_handler = ProcessSafeHandler(stream=sys.stdout)
    console_handler.setLevel(console_verbose_level)
    console_formatter = ConsoleFormatter()
    console_handler.setFormatter(console_formatter)
    lgr.addHandler(console_handler)
    lgr.propagate = False

    # time str to be used by filename
    timestr = time.strftime("%d_%m_%y__%H_%M")

    # log to files - only if output_path is provided
    if output_path is not None:
        log_output_path = os.path.join(output_path, "logs")
        # Create log dir
        create_dir(log_output_path)

        # file info
        file_handler = logging.FileHandler(os.path.join(log_output_path, f"fuse_{timestr}.log"))
        file_handler.setLevel(logging.INFO)
        formatter = logging.Formatter("(%(filename)s:%(lineno)d) : %(levelname)s : %(message)s")
        file_handler.setFormatter(formatter)
        lgr.addHandler(file_handler)

        # file verbose
        file_handler = RotatingFileHandler(os.path.join(log_output_path, "fuse_verbose.log"), maxBytes=1e6)
        file_handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(asctime)s : (%(filename)s:%(lineno)d) : %(message)s")
        file_handler.setFormatter(formatter)
        lgr.addHandler(file_handler)

    start_time = time.strftime("%d/%m/%y %H:%M:%S")
    lgr.info(f"Fuse: {start_time}", {"color": "cyan", "attrs": ["dark", "bold"]})

    if list_of_source_files is None or list_of_source_files:
        if output_path is not None:
            source_files_output_path = os.path.join(output_path, "source_files", timestr)
            create_or_reset_dir(source_files_output_path)

            lgr.info(f"Copy source files to {source_files_output_path}")
            if list_of_source_files is None:
                # copy just the caller function file name
                caller_function_file_name = inspect.stack()[1][1]
                list_of_source_files = [caller_function_file_name]

            for src_file in list_of_source_files:
                copyfile(os.path.abspath(src_file), os.path.join(source_files_output_path, os.path.basename(src_file)))


def fuse_logger_end():
    lgr = logging.getLogger("Fuse")
    # skip if not already configured
    if lgr.level == 0:
        return
    for h in lgr.handlers:
        lgr.removeHandler(h)


def convert_state_to_str(input_state: Any):
    """
    Convert a state built from a nested dictionaries, lists and tuples to a string
    :param input_state:
    :return:
    """
    if isinstance(input_state, dict):
        return {param_name: convert_state_to_str(input_state[param_name]) for param_name in input_state}
    if isinstance(input_state, list):
        return [convert_state_to_str(param) for param in input_state]
    if isinstance(input_state, tuple):
        return tuple((convert_state_to_str(param) for param in input_state))
    return str(input_state)


def log_object_input_state(obj: Any, input_state: Dict) -> None:
    """
    Log object creation: module name, address and input parameters
    :param obj: The object, typically set to 'self'.
    :param input_state: The input parameters,
                             To get them call this function first thing in the constructors with input_parameters=locals()
    :return: None
    """
    lgr = logging.getLogger("Fuse")
    input_state_copy = input_state.copy()
    # remove self from the state if exist
    input_state_copy.pop("self", None)
    input_paramters_str = convert_state_to_str(input_state_copy)
    lgr.debug(f"\n-----\n{str(obj)} created: \n\tinput state {input_paramters_str} \n-----\n")



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, List

import torch
import torch.nn as nn


class BackboneMultilayerPerceptron(torch.nn.Module):
    def __init__(
        self,
        layers: List[int] = (64, 192, 320, 320, 1088, 384),
        mlp_input_size: int = 103,
        activation_layer: torch.nn.Module = nn.ReLU(inplace=False),
        dropout_rate: float = 0.0,
    ) -> None:
        super().__init__()

        mlp_layers = [nn.Linear(mlp_input_size, layers[0])]
        if activation_layer is not None:
            mlp_layers.append(activation_layer)
        if dropout_rate > 0:
            mlp_layers.append(nn.Dropout(p=dropout_rate))

        for layer_idx in range(len(layers) - 1):
            mlp_layers.append(nn.Linear(layers[layer_idx], layers[layer_idx + 1]))
            mlp_layers.append(activation_layer) if activation_layer is not None else None
            mlp_layers.append(nn.Dropout(p=dropout_rate)) if dropout_rate > 0 else None

        self.mlp = nn.ModuleList(mlp_layers)

    def forward(self, input_tensor: Dict) -> Dict:
        for layer in self.mlp:
            input_tensor = layer(input_tensor)
        return input_tensor



import os
from fuse.data.ops.op_base import OpReversibleBase
from typing import Optional
import numpy as np
import nibabel as nib
from fuse.utils.ndict import NDict

from torchvision.io import read_image
from medpy.io import load
import pydicom


class OpLoadImage(OpReversibleBase):
    """
    Loads a medical image, currently supports:
            'nii', 'nib', 'jpg', 'jpeg', 'png', 'mha','dcm'
    """

    def __init__(self, dir_path: str, **kwargs):
        super().__init__(**kwargs)
        self._dir_path = dir_path

    def __call__(
        self,
        sample_dict: NDict,
        op_id: Optional[str],
        key_in: str,
        key_out: str,
        key_metadata_out: Optional[str] = None,
        format: str = "infer",
    ):
        """
        :param key_in: the key name in sample_dict that holds the filename
        :param key_out: the key name in sample_dict that holds the image
        :param key_metadata_out : the key to hold metadata dictionary
        """
        img_filename = os.path.join(self._dir_path, sample_dict[key_in])
        img_filename_suffix = img_filename.split(".")[-1]
        if (format == "infer" and img_filename_suffix in ["nii"]) or (format in ["nii", "nib"]):

            img = nib.load(img_filename)
            img_np = img.get_fdata()
            sample_dict[key_out] = img_np

        elif img_filename_suffix in ["jpg", "jpeg", "png"]:
            img = read_image(img_filename)
            img = img.float()
            img_np = img.numpy()
            sample_dict[key_out] = img_np

        elif (format == "infer" and img_filename_suffix in ["mha"]) or (format in ["mha"]):
            image_data, image_header = load(img_filename)
            sample_dict[key_out] = image_data
            if key_metadata_out is not None:
                sample_dict[key_metadata_out] = {
                    key: image_header.sitkimage.GetMetaData(key) for key in image_header.sitkimage.GetMetaDataKeys()
                }
        elif (format == "infer" and img_filename_suffix in ["dcm"]) or (format in ["dcm"]):
            dcm = pydicom.dcmread(img_filename)
            inner_image = dcm.pixel_array
            # convert to numpy
            img_np = np.asarray(inner_image)
            sample_dict[key_out] = img_np
            if key_metadata_out is not None:
                metadata = NDict()
                for key, field in key_metadata_out:
                    metadata[field] = dcm[key].value
                sample_dict[key_metadata_out] = metadata
        else:
            raise Exception(
                f"OpLoadImage: case format {format} and {img_filename_suffix} is not supported - filename {img_filename}"
            )

        return sample_dict

    def reverse(self, sample_dict: dict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        return sample_dict



"""
(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
   http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Created on June 30, 2021
"""
from functools import partial
from typing import Hashable, List, Optional, Sequence, Union, Callable, Any, Tuple

from fuse.data.pipelines.pipeline_default import PipelineDefault
from collections import OrderedDict
from fuse.data.datasets.caching.object_caching_handlers import _object_requires_hdf5_recurse
from fuse.utils.ndict import NDict
import os
import psutil
from fuse.utils.file_io.file_io import load_hdf5, save_hdf5_safe, load_pickle, save_pickle_safe
from fuse.data import get_sample_id, create_initial_sample, get_specific_sample_from_potentially_morphed
import hashlib
from fuse.utils.file_io import delete_directory_tree
from glob import glob
from fuse.utils.multiprocessing.run_multiprocessed import run_multiprocessed, get_from_global_storage
from fuse.data.datasets.sample_caching_audit import SampleCachingAudit
from fuse.data.utils.sample import get_initial_sample_id, set_initial_sample_id
from warnings import warn


class SamplesCacher:
    def __init__(
        self,
        unique_name: str,
        pipeline: PipelineDefault,
        cache_dirs: Union[str, List[str]],
        custom_write_dir_callable: Optional[Callable] = None,
        custom_read_dirs_callable: Optional[Callable] = None,
        restart_cache: bool = False,
        workers: int = 0,
        verbose=1,
        **audit_kwargs: dict,
    ) -> None:
        """
        Supports caching samples, used by datasets implementations.
        :param unique_name: a unique name for this cache.
         cache dir will be [cache dir]/[unique_name]
        :param cache_dirs: a path in which the cache will be created,
         you may provide a list of paths, which will be tried in order, moving the next when available space is exausted.
        :param parameter:
        :param custom_write_dir_callable: optional callable with the signature foo(cache_dirs:List[str]) -> str
         which returns the write directory to use.
        :param custom_read_dirs_callable: optional callable with the signature foo() -> List[str]
         which returns a list of directories to attempt to read from. Attempts will be in the provided order.
        :param restart_cache: if set to True, will DELETE all of the content of the defined cache dirs.
        Should be used every time that any of the OPs participating in the "static cache" part changed in any way
        (for example, code change)
        :param workers: number of multiprocessing workers used when building the cache. Default value is 0 (no multiprocessing)
        :param **audit_kwargs: optional custom kwargs to pass to SampleCachingAudit instance.
            auditing cached samples (usually periodically) is very important, in order to avoid "stale" cached samples.
            To disable pass audit_first_sample=False, audit_rate=None,
            Note that it's not recommended to completely disable it, and at the very least you should use audit_first_sample=True, audit_rate=None
            which only tests the first loaded sample for staleness.
            To learn more read SampleCachingAudit doc
        """
        if not isinstance(cache_dirs, list):
            cache_dirs = [cache_dirs]
        self._cache_dirs = [os.path.join(x, unique_name) for x in cache_dirs]

        self._unique_name = unique_name

        if custom_write_dir_callable is None:
            self._write_dir_logic = _get_available_write_location
        else:
            self._write_dir_logic = custom_write_dir_callable

        if custom_read_dirs_callable is None:
            self._read_dirs_logic = partial(default_read_dirs_logic, cache_dirs=self._cache_dirs)
        else:
            self._read_dirs_logic = custom_read_dirs_callable

        self._pipeline = pipeline
        self._pipeline_desc_text = str(pipeline)
        self._pipeline_desc_hash = "hash_" + hashlib.md5(self._pipeline_desc_text.encode("utf-8")).hexdigest()

        self._verbose = verbose

        if self._verbose > 0:
            print(f"pipeline description hash for [{unique_name}] is: {self._pipeline_desc_hash}")

        self._restart_cache = restart_cache
        if self._restart_cache:
            self.delete_cache()

        self._audit_kwargs = audit_kwargs
        self._audit = SampleCachingAudit(**self._audit_kwargs)

        self._workers = workers
        if self._workers < 2:
            warn(
                'Multi processing is not active in SamplesCacher. Seting "workers" to the number of your cores usually results in a significant speedup. Debugging, however, is easier with "workers=0".'
            )

        self._verify_no_other_pipelines_cache()

    def _verify_no_other_pipelines_cache(self) -> None:
        dirs_to_check = self._get_read_dirs() + [self._get_write_dir()]
        for d in dirs_to_check:
            search_pat = os.path.realpath(os.path.join(d, "..", "hash_*"))
            found_sub_dirs = glob(search_pat)
            for found_dir in found_sub_dirs:
                if not os.path.isdir(found_dir):
                    continue
                if os.path.basename(found_dir) != self._pipeline_desc_hash:
                    raise Exception(
                        f"Found samples cache for pipeline hash {os.path.basename(found_dir)} which is different from the current loaded pipeline hash {self._pipeline_desc_hash} !!\n"
                        "This is not allowed, you may only use a single pipeline per uniquely named cache.\n"
                        'You can use "restart_cache=True" to rebuild the cache or delete the different cache manually.\n'
                    )

    def delete_cache(self) -> None:
        """
        Will delete this specific named cache from all read and write dirs
        """
        dirs_to_delete = self._get_read_dirs() + [self._get_write_dir()]
        dirs_to_delete = list(set(dirs_to_delete))
        dirs_to_delete = [
            os.path.realpath(os.path.join(x, "..")) for x in dirs_to_delete
        ]  # one dir above the pipeline hash dir
        print('Due to "delete_cache" call, about to delete the following dirs:')

        for del_dir in dirs_to_delete:
            print(del_dir)
        print("---- list end ----")
        print("deleting ... ")
        for del_dir in dirs_to_delete:
            print(f"deleting {del_dir} ...")
            all_found = glob(os.path.join(del_dir, "hash_*"))
            for found in all_found:
                if not os.path.isdir(found):
                    continue
                delete_directory_tree(found)

    def _get_write_dir(self):
        ans = self._write_dir_logic(self._cache_dirs)
        ans = os.path.join(ans, self._pipeline_desc_hash)
        return ans

    def _get_read_dirs(self):
        ans = self._read_dirs_logic()
        ans = [os.path.join(x, self._pipeline_desc_hash) for x in ans]
        return ans

    def cache_samples(self, orig_sample_ids: List[Any]) -> List[Tuple[str, Union[None, List[str]], str]]:
        """
        Go over all of orig_sample_ids, and cache resulting samples
        returns information that helps to map from original sample id to the resulting sample id
        (an op might return None, discarding a sample, or optional generate different one or more samples from an original single sample_id)
        #TODO: have a single doc location that explains this concept and can be pointed to from any related location
        """
        # TODO: remember that it means that we need proper extraction of args (pos or kwargs...)
        # possibly by extracting info from __call__ signature or process() if we modify from call to it

        # TODO:

        sample_ids_text = "@".join([str(x) for x in sorted(orig_sample_ids)])
        samples_ids_hash = hashlib.md5(sample_ids_text.encode("utf-8")).hexdigest()

        hash_filename = "samples_ids_hash@" + samples_ids_hash + ".pkl.gz"

        read_dirs = self._get_read_dirs()
        for curr_read_dir in read_dirs:
            fullpath_filename = os.path.join(curr_read_dir, "full_sets_info", hash_filename)
            if os.path.isfile(fullpath_filename):
                print(f"entire samples set {hash_filename} already cached. Found {fullpath_filename}")
                return load_pickle(fullpath_filename)

        orig_sid_to_final = OrderedDict()
        for_global_storage = {"samples_cacher_instance": self}
        all_ans = run_multiprocessed(
            SamplesCacher._cache_worker,
            orig_sample_ids,
            workers=self._workers,
            copy_to_global_storage=for_global_storage,
            verbose=1,
        )

        for initial_sample_id, output_sample_ids in zip(orig_sample_ids, all_ans):
            orig_sid_to_final[initial_sample_id] = output_sample_ids

        write_dir = self._get_write_dir()
        set_info_dir = os.path.join(write_dir, "full_sets_info")
        os.makedirs(set_info_dir, exist_ok=True)
        fullpath_filename = os.path.join(set_info_dir, hash_filename)
        save_pickle_safe(orig_sid_to_final, fullpath_filename, compress=True)

        return orig_sid_to_final

    @staticmethod
    def get_final_sample_id_hash(sample_id):
        """
        sample_id is the final sample_id that came out of the pipeline
        note: our pipeline supports Ops returning None, thus, discarding a sample (in that case, it will not have any final sample_id),
        additionally, the pipeline may return *multiple* samples, each with their own sample_id
        """
        curr_sample_id_str = str(sample_id)  # TODO repr or str ?
        output_sample_hash = hashlib.md5(curr_sample_id_str.encode("utf-8")).hexdigest()
        ans = f"out_sample_id@{output_sample_hash}"
        return ans

    @staticmethod
    def get_orig_sample_id_hash(orig_sample_id):
        """
        orig_sample_id is the original sample_id that was provided, regardless if it turned out to become None, the same sample_id, or different sample_id(s)
        """
        orig_sample_id_str = str(orig_sample_id)
        if orig_sample_id_str.startswith("<") and orig_sample_id_str.endswith(">"):  # and '0x' in orig_sample_id_str
            # <__main__.SomeClass at 0x7fc3e6645e20>
            raise Exception(
                f"You must implement a proper __str__ for orig_sample_id. String representations like <__main__.SomeClass at 0x7fc3e6645e20> are not descriptibe enough and also not persistent between runs. Got: {orig_sample_id_str}"
            )
        ans = hashlib.md5(orig_sample_id_str.encode("utf-8")).hexdigest()
        ans = "out_info_for_orig_sample@" + ans
        return ans

    def get_orig_sample_id_from_final_sample_id(self, orig_sample_id):
        pass

    def load_sample(self, sample_id: Hashable, keys: Optional[Sequence[str]] = None):
        """
        :param sample_id: the sample_id of the sample to load
        :param keys: optionally, provide a subset of the keys to load in this sample.
        This is useful for speeding up loading.
        """

        sample_from_cache = self._load_sample_from_cache(sample_id, keys)
        audit_required = self._audit.update()

        if audit_required:
            initial_sample_id = get_initial_sample_id(sample_from_cache)
            fresh_sample = self._load_sample_using_pipeline(initial_sample_id, keys)
            fresh_sample = get_specific_sample_from_potentially_morphed(fresh_sample, sample_id)

            self._audit.audit(sample_from_cache, fresh_sample)

        return sample_from_cache

    def _load_sample_using_pipeline(self, sample_id: Hashable, keys: Optional[Sequence[str]] = None):
        sample_dict = create_initial_sample(sample_id)
        result_sample = self._pipeline(sample_dict)
        return result_sample

    def _load_sample_from_cache(self, sample_id: Hashable, keys: Optional[Sequence[str]] = None):
        """
        TODO: add comments
        """
        read_dirs = self._get_read_dirs()
        sample_hash = SamplesCacher.get_final_sample_id_hash(sample_id)

        for curr_read_dir in read_dirs:
            extension_less = os.path.join(curr_read_dir, sample_hash)
            if os.path.isfile(extension_less + ".pkl.gz"):
                loaded_sample = NDict(load_pickle(extension_less + ".pkl.gz"))
                if os.path.isfile(extension_less + ".hdf5"):
                    loaded_sample_hdf5_part = load_hdf5(extension_less + ".hdf5")
                    loaded_sample.merge(loaded_sample_hdf5_part)
                return loaded_sample

        raise Exception(f"Expected to find a cached sample for sample_id={sample_id} but could not find any!")

    @staticmethod
    def _cache_worker(orig_sample_id: Any):
        cacher = get_from_global_storage("samples_cacher_instance")
        ans = cacher._cache(orig_sample_id)
        return ans

    def _cache(self, orig_sample_id: Any):
        """
        :param orig_sample_id: the original sample id, which was provided as the input to the pipeline
        :param sample: the result of the pipeline - can be None if it was dropped, a dictionary in the typical standard case,
         and a list of dictionaries in case the sample was split into multiple samples (ops are allowed to do that during the static part of the processing)
        """

        write_dir = self._get_write_dir()
        os.makedirs(write_dir, exist_ok=True)
        read_dirs = self._get_read_dirs()

        was_processed_hash = SamplesCacher.get_orig_sample_id_hash(orig_sample_id)
        was_processed_fn = was_processed_hash + ".pkl"

        # checking in all read directories if information related to this sample(s) was already cached
        for curr_read_dir in read_dirs:
            fn = os.path.join(curr_read_dir, was_processed_fn)
            if os.path.isfile(fn):
                ans = load_pickle(fn)
                return ans

        result_sample = self._load_sample_using_pipeline(orig_sample_id)

        if isinstance(result_sample, dict):
            result_sample = [result_sample]

        if isinstance(result_sample, list):
            if 0 == len(result_sample):
                result_sample = None
            for s in result_sample:
                set_initial_sample_id(s, orig_sample_id)

        if not isinstance(result_sample, (list, dict, type(None))):
            raise Exception(
                f"Unsupported sample type, got {type(result_sample)}. Supported types are dict, list-of-dicts and None."
            )

        if result_sample is not None:
            output_info = []
            for curr_sample in result_sample:
                curr_sample_id = get_sample_id(curr_sample)
                output_info.append(curr_sample_id)
                output_sample_hash = SamplesCacher.get_final_sample_id_hash(curr_sample_id)

                requiring_hdf5_keys = _object_requires_hdf5_recurse(curr_sample)
                if len(requiring_hdf5_keys) > 0:
                    requiring_hdf5_dict = curr_sample.get_multi(requiring_hdf5_keys)
                    requiring_hdf5_dict = requiring_hdf5_dict.flatten()

                    hdf5_filename = os.path.join(write_dir, output_sample_hash + ".hdf5")
                    save_hdf5_safe(hdf5_filename, **requiring_hdf5_dict)

                    # remove all hdf5 entries from the sample_dict that will be pickled
                    for k in requiring_hdf5_dict:
                        _ = curr_sample.pop(k)

                save_pickle_safe(curr_sample, os.path.join(write_dir, output_sample_hash + ".pkl.gz"), compress=True)
        else:
            output_info = None
            # requiring_hdf5_keys = None

        save_pickle_safe(output_info, os.path.join(write_dir, was_processed_fn))
        return output_info


def _get_available_write_location(cache_dirs: List[str], max_allowed_used_space=0.95):
    """
    :param cache_dirs: write directories. Directories are checked in order that they are provided.
    :param max_allowed_used_space: set to a value between 0.0 to 1.0.
    a value of 0.95 means that once the available space is greater or equal to 95% of the the disk capacity,
    it will be considered full, and the next directory will be attempted.
    """

    for curr_loc in cache_dirs:
        if max_allowed_used_space is None:
            return curr_loc
        os.makedirs(curr_loc, exist_ok=True)
        drive_stats = psutil.disk_usage(curr_loc)
        actual_usage_part = drive_stats.percent / 100.0
        if actual_usage_part < max_allowed_used_space:
            return curr_loc

    raise Exception(
        "Could not find any location to write.\n"
        f"write_cache_locations={cache_dirs}\n"
        f"max_allowed_used_space={max_allowed_used_space}"
    )


def default_read_dirs_logic(cache_dirs: List[str]):
    return cache_dirs



import torch
import numpy as np
import random


class Seed:
    """
    Random seed functionality - static methods
    """

    @staticmethod
    def set_seed(seed: int, deterministic_mode: bool = True) -> torch.Generator:
        """
        Fix a seed (numpy, torch and random) and set PyTorch mode to deterministic to reproduce results
        :param seed: seed to use
        :param deterministic_mode: set PyTorch deterministic mode
        :return: Random generator to be used when creating Dataloader
        """
        # random
        random.seed(seed)

        # torch
        torch.manual_seed(seed)
        torch.use_deterministic_algorithms(deterministic_mode)
        dataloader_rand_gen = torch.Generator()
        dataloader_rand_gen.manual_seed(seed)

        # numpy
        np.random.seed(seed)

        return dataloader_rand_gen

    @staticmethod
    def seed_worker_init(worker_id: int):
        """
        Function to provide to torch dataloader to set a seed per worker
        """
        worker_seed = torch.initial_seed() % 2**32
        Seed.set_seed(worker_seed)



import unittest

from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuseimg.data.ops.color import OpClip, OpToRange
from fuseimg.data.ops.shape_ops import OpPad

from fuse.utils.ndict import NDict

import numpy as np
import torch


class TestOps(unittest.TestCase):
    def test_basic_1(self):
        """
        Test basic imaging ops
        """

        sample = NDict()
        sample["data.input.img"] = np.array([5, 0.5, -5, 3])

        pipeline = PipelineDefault(
            "test_pipeline",
            [
                # (op_normalize_against_self, {} ),
                (OpClip(), dict(key="data.input.img", clip=(-0.5, 3.0))),
                (OpToRange(), dict(key="data.input.img", from_range=(-0.5, 3.0), to_range=(-3.5, 3.5))),
            ],
        )

        sample = pipeline(sample)

        self.assertLessEqual(sample["data.input.img"].max(), 3.5)
        self.assertGreaterEqual(sample["data.input.img"].min(), -3.5)
        self.assertEqual(sample["data.input.img"][-1], 3.5)

    def test_op_pad(self):
        """
        Test OpPad
        """
        sample = NDict()
        sample["data.input.tensor_img_1"] = torch.Tensor([[1]])
        sample["data.input.numpy_img_1"] = np.array([[1]])
        sample["data.input.tensor_img_2"] = torch.Tensor([[42]])
        sample["data.input.numpy_img_2"] = np.array([[42]])

        pipeline = PipelineDefault(
            "test_pipeline",
            [
                (OpPad(), dict(key="data.input.tensor_img_1", padding=1, fill=0, mode="constant")),
                (OpPad(), dict(key="data.input.numpy_img_1", padding=1, fill=0, mode="constant")),
                (OpPad(), dict(key="data.input.tensor_img_2", padding=1, fill=42, mode="constant")),
                (OpPad(), dict(key="data.input.numpy_img_2", padding=1, fill=42, mode="constant")),
            ],
        )

        pipeline(sample)

        res_1 = [[0, 0, 0], [0, 1, 0], [0, 0, 0]]

        res_2 = [[42, 42, 42], [42, 42, 42], [42, 42, 42]]

        self.assertTrue(np.array_equal(sample["data.input.tensor_img_1"], res_1))
        self.assertTrue(np.array_equal(sample["data.input.numpy_img_1"], res_1))
        self.assertTrue(np.array_equal(sample["data.input.tensor_img_2"], res_2))
        self.assertTrue(np.array_equal(sample["data.input.numpy_img_2"], res_2))

    def test_op_resize_to(self):
        pass


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict, Sequence
from fuse.data.patterns import Patterns


class DataTypeBasic(Enum):
    UNKNOWN = -1  # TODO: change to Unknown?


class TypeDetectorBase(ABC):
    @abstractmethod
    def get_type(self, sample_dict: Dict, key: str):
        """
        Returns the type of key
        The most common implementation can be seen in TypeDetectorPatternsBased.
        """
        raise NotImplementedError

    @abstractmethod
    def verify_type(self, sample_dict: Dict, key: str, types: Sequence[Enum]):
        """
        Raises exception if key is not one of the types found in types
        """
        raise NotImplementedError


class TypeDetectorPatternsBased(TypeDetectorBase):
    def __init__(self, patterns_dict: Dict[str, Enum]):
        """
        type detection based on the key (NDict "style" - for example 'data.cc.img')
        get_type ignores the sample_dict completely.
        TODO: provide usage example
        """
        self._patterns_dict = patterns_dict
        self._patterns = Patterns(self._patterns_dict, DataTypeBasic.UNKNOWN)

    def get_type(self, sample_dict: Dict, key: str):
        return self._patterns.get_value(key)

    def verify_type(self, sample_dict: Dict, key: str, types: Sequence[Enum]):
        self._patterns.verify_value_in(key, types)



import os
import pathlib
import shutil
from tempfile import mkdtemp
import unittest
from fuse.data.utils.sample import get_sample_id
from fuse.utils.file_io.file_io import create_dir

from fuseimg.datasets.kits21 import KITS21
from fuseimg.datasets.stoic21 import STOIC21
from tqdm import trange
from testbook import testbook
from fuse.eval.metrics.stat.metrics_stat_common import MetricUniqueValues
from fuse.utils.multiprocessing.run_multiprocessed import get_from_global_storage, run_multiprocessed
from fuse.eval.evaluator import EvaluatorDefault

from fuseimg.datasets.isic import ISIC
from examples.fuse_examples.imaging.classification.isic.golden_members import FULL_GOLDEN_MEMBERS, TEN_GOLDEN_MEMBERS


notebook_path = os.path.join(pathlib.Path(__file__).parent.resolve(), "../kits21_example.ipynb")


def ds_getitem(index: int):
    sample = get_from_global_storage("ds")[index].flatten()
    del sample["data.input.img"]
    return sample


class TestDatasets(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.kits21_cache_dir = mkdtemp(prefix="kits21_cache")
        self.kits21_data_dir = (
            os.environ["KITS21_DATA_PATH"] if "KITS21_DATA_PATH" in os.environ else mkdtemp(prefix="kits21_data")
        )

        self.stoic21_cache_dir = mkdtemp(prefix="stoic_cache")

        self.isic_cache_dir = mkdtemp(prefix="isic_cache")
        self.isic_data_dir = (
            os.environ["ISIC19_DATA_PATH"] if "ISIC19_DATA_PATH" in os.environ else mkdtemp(prefix="isic_data")
        )

    def test_kits21(self):
        KITS21.download(self.kits21_data_dir, cases=list(range(10)))

        create_dir(self.kits21_cache_dir)
        dataset = KITS21.dataset(
            data_path=self.kits21_data_dir,
            cache_dir=self.kits21_cache_dir,
            reset_cache=True,
            sample_ids=[f"case_{id:05d}" for id in range(10)],
        )
        self.assertEqual(len(dataset), 10)
        for sample_index in trange(10):
            sample = dataset[sample_index]
            self.assertEqual(get_sample_id(sample), f"case_{sample_index:05d}")

    @unittest.skipIf(
        "STOIC21_DATA_PATH" not in os.environ, "Expecting environment variable STOIC21_DATA_PATH to be defined"
    )
    def test_stoic21(self):
        data_path = os.environ["STOIC21_DATA_PATH"]
        sids = STOIC21.sample_ids(data_path)[:10]
        ds = STOIC21.dataset(sample_ids=sids, data_path=data_path, cache_dir=self.stoic21_cache_dir, reset_cache=True)

        metrics = {
            "age": MetricUniqueValues(key="data.input.age"),
            "gender": MetricUniqueValues(key="data.input.gender"),
            "thickness": MetricUniqueValues(key="data.metadata.SliceThickness"),
            "covid": MetricUniqueValues(key="data.gt.probCOVID"),
            "severe": MetricUniqueValues(key="data.gt.probSevere"),
        }
        evaluator = EvaluatorDefault()

        data_iter = run_multiprocessed(
            worker_func=ds_getitem,
            args_list=range(len(ds)),
            copy_to_global_storage={"ds": ds},
            workers=10,
            verbose=1,
            as_iterator=True,
        )

        results = evaluator.eval(ids=None, data=data_iter, metrics=metrics, id_key="data.sample_id")

        self.assertEqual(ds[0]["data.input.clinical"].shape[0], 8)
        self.assertTrue(5 in dict(results["metrics.age"]))

    def test_isic(self):

        create_dir(self.isic_cache_dir)
        dataset = ISIC.dataset(
            self.isic_data_dir, self.isic_cache_dir, reset_cache=True, samples_ids=TEN_GOLDEN_MEMBERS
        )
        self.assertEqual(len(dataset), 10)
        for sample_index in range(10):
            sample = dataset[sample_index]
            self.assertEqual(get_sample_id(sample), TEN_GOLDEN_MEMBERS[sample_index])

    @testbook(notebook_path, execute=range(0, 4), timeout=120)
    def test_basic(tb, self):
        tb.execute_cell([4, 5])

        tb.inject(
            """
            assert(np.max(my_dataset[0]['data.input.img'])>=0 and np.max(my_dataset[0]['data.input.img'])<=1)
            """
        )

    @testbook(notebook_path, execute=range(0, 4), timeout=240)
    def test_caching(tb, self):
        tb.execute_cell([9])

        tb.execute_cell([16, 17])
        tb.inject(
            """
            assert(isinstance(my_dataset[0]["data.gt.seg"], torch.Tensor))
            """
        )

    @testbook(notebook_path, execute=range(0, 4), timeout=240)
    def test_custom(tb, self):
        tb.execute_cell([25])

        tb.inject(
            """
            assert(my_dataset[0]["data.gt.seg"].shape[1:] == (4, 256, 256))
            """
        )

    def tearDown(self) -> None:
        shutil.rmtree(self.kits21_cache_dir)
        if "KITS21_DATA_PATH" not in os.environ:
            shutil.rmtree(self.kits21_data_dir)

        shutil.rmtree(self.stoic21_cache_dir)

        shutil.rmtree(self.isic_cache_dir)
        if "ISIC19_DATA_PATH" not in os.environ:
            shutil.rmtree(self.isic_data_dir)

        super().tearDown()


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import shutil
import tempfile
import unittest
import os
from unittest import result
from fuse.utils.multiprocessing.run_multiprocessed import run_in_subprocess
from fuse.utils.ndict import NDict

from fuse.utils.rand.seed import Seed

from fuse_examples.imaging.classification.mnist.run_mnist import (
    EVAL_COMMON_PARAMS,
    INFER_COMMON_PARAMS,
    TRAIN_COMMON_PARAMS,
    run_train,
    run_infer,
    run_eval,
)


def run_mnist(root: str) -> None:
    model_dir = os.path.join(root, "model_dir")
    paths = {
        "model_dir": model_dir,
        "force_reset_model_dir": True,  # If True will reset model dir automatically - otherwise will prompt 'are you sure' message.
        "cache_dir": os.path.join(root, "cache_dir"),
        "inference_dir": os.path.join(model_dir, "infer_dir"),
        "eval_dir": os.path.join(model_dir, "eval_dir"),
    }

    train_common_params = TRAIN_COMMON_PARAMS

    infer_common_params = INFER_COMMON_PARAMS

    eval_common_params = EVAL_COMMON_PARAMS
    Seed.set_seed(0, False)  # previous test (in the pipeline) changed the deterministic behavior to True
    run_train(paths, train_common_params)
    run_infer(paths, infer_common_params)
    results = run_eval(paths, eval_common_params)
    assert results["metrics.auc.macro_avg"] >= 0.95, "Error: expecting higher performence"


class ClassificationMnistTestCase(unittest.TestCase):
    def setUp(self):
        self.root = tempfile.mkdtemp()

    def test_template(self):
        run_in_subprocess(run_mnist, self.root)

    def tearDown(self):
        # Delete temporary directories
        shutil.rmtree(self.root)


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import torch
import torch.nn as nn
import numpy as np
from fuse.dl.losses.loss_base import LossBase
from fuse.utils.ndict import NDict
from typing import Callable, Optional


def make_one_hot(input, num_classes):
    """Convert class index tensor to one hot encoding tensor.
    Args:
         input: A tensor of shape [N, 1, *]
         num_classes: An int of number of class
    Returns:
        A tensor of shape [N, num_classes, *]
    """
    shape = np.array(input.shape)
    shape[1] = num_classes
    shape = tuple(shape)
    result = torch.zeros(shape, device=input.device)
    result = result.scatter_(1, input, 1)

    return result


class BinaryDiceLoss(nn.Module):
    def __init__(self, power: int = 1, eps: float = 1.0, reduction: str = "mean"):
        """
        :param power:       Denominator value: \sum{x^p} + \sum{y^p}, default: 1
        :param eps:         A float number to smooth loss, and avoid NaN error, default: 1
        :param reduction:   Reduction method to apply, return mean over batch if 'mean',
                            return sum if 'sum', return a tensor of shape [N,] if 'none'

        Returns:            Loss tensor according to arg reduction
        Raise:              Exception if unexpected reduction
        """
        super().__init__()
        self.p = power
        self.reduction = reduction
        self.eps = eps

    def __call__(self, predict, target):
        assert predict.shape[0] == target.shape[0], "predict & target batch size don't match"
        predict = predict.contiguous().view(predict.shape[0], -1)
        target = target.contiguous().view(target.shape[0], -1)

        if target.dtype == torch.int64:
            target = target.type(torch.float32).to(target.device)
        num = 2 * torch.sum(torch.mul(predict, target), dim=1) + self.eps
        den = torch.sum(predict.pow(self.p) + target.pow(self.p), dim=1) + self.eps
        loss = 1 - num / den

        # return loss
        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        elif self.reduction == "none":
            return loss
        else:
            raise Exception("Unexpected reduction {}".format(self.reduction))


class DiceLoss(LossBase):
    def __init__(
        self,
        pred_name,
        target_name,
        filter_func: Optional[Callable] = None,
        class_weights=None,
        ignore_cls_index_list=None,
        resize_mode: str = "maxpool",
        **kwargs
    ):
        """

        :param pred_name:                batch_dict key for predicted output (e.g., class probabilities after softmax).
                                         Expected Tensor shape = [batch, num_classes, height, width]
        :param target_name:              batch_dict key for target (e.g., ground truth label). Expected Tensor shape = [batch, height, width]
        :param filter_func:              function that filters batch_dict/ The function gets ans input batch_dict and returns filtered batch_dict
        :param class_weights:            An array of shape [num_classes,]
        :param ignore_cls_index_list:    class index to ignore (list)
        :param resize_mode:              Resize mode- either using a max pooling kernel(default), or using PyTorch
                                         interpolation ('interpolate'/'maxpool')
        :param kwargs:                   args pass to BinaryDiceLoss
        """

        super().__init__(pred_name, target_name, class_weights)
        self.class_weights = class_weights
        self.filter_func = filter_func
        self.kwargs = kwargs
        self.ignore_cls_index_list = ignore_cls_index_list
        self.resize_mode = resize_mode
        self.dice = BinaryDiceLoss(**self.kwargs)

    def __call__(self, batch_dict: NDict):

        if self.filter_func is not None:
            batch_dict = self.filter_func(batch_dict)
        predict = batch_dict[self.pred_name].float()
        target = batch_dict[self.target_name].long()

        n, c, h, w = predict.shape
        tar_shape = target.shape
        if len(tar_shape) < 4:
            target = target.unsqueeze(1)
        nt, ct, ht, wt = target.shape

        if h != ht or w != wt:  # upsample
            if self.resize_mode == "maxpool":
                block_height = int(ht / h)
                block_width = int(wt / w)
                residual_h = int((ht - (block_height * h)) / 2)
                residual_w = int((wt - (block_width * w)) / 2)

                target = torch.nn.functional.max_pool2d(
                    target[:, :, residual_h : ht - residual_h, residual_w : wt - residual_w],
                    kernel_size=(block_height, block_width),
                )
            elif self.resize_mode == "interpolate":
                target = torch.nn.functional.interpolate(target, size=(h, w))
            else:
                raise Exception

        total_loss = 0
        n_classes = predict.shape[1]

        # Convert target to one hot encoding
        if n_classes > 1 and target.shape[1] != n_classes:
            target = make_one_hot(target, n_classes)

        assert predict.shape == target.shape, "predict & target shape do not match"

        total_class_weights = sum(self.class_weights) if self.class_weights is not None else n_classes
        for cls_index in range(n_classes):
            if cls_index not in self.ignore_cls_index_list:
                dice_loss = self.dice(predict[:, cls_index, :, :], target[:, cls_index, :, :])
                if self.class_weights is not None:
                    assert self.class_weights.shape[0] == n_classes, "Expect weight shape [{}], got[{}]".format(
                        n_classes, self.class_weights.shape[0]
                    )
                    dice_loss *= self.class_weights[cls_index]
                total_loss += dice_loss
        total_loss /= total_class_weights

        return self.weight * total_loss



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from abc import abstractmethod
from fuse.utils.ndict import NDict
import torch


class LossBase(torch.nn.Module):
    """
    Base class for Fuse loss functions.
    Essentially it is torch.nn.Module whose forward method gets as an input batch_dict and returns loss tensor
    Note - if you already have a pytorch implementation of the loss, a useful alternative that you can use is LossDefault class to wrap it

    LossBase Usage example:

    '''
    from fuse.dl.losses import LossBase
    Class MyCustomLoss(LossBase):
          '''
          Extracts prediction and data from batch_dict and performs a dummy loss calculation,
          clamping the loss value to be at most "cap"
          '''
          def __init__(self,
             pred: str = None,
             target: str = None,
             cap:float = 0.3,
             ):
             self._pred = pred
             self._target = target
             self._cap = cap

          def forward(self, batch_dict:NDict) -> torch.Tensor:
             #extract pred and target from batch_dict
             pred = batch_dict[self.pred]
             target = batch_dict[self.target]

             #apply our loss logic, and clamp to cap
             z = torch.mean(pred)-torch.mean(target)
             z = torch.sqrt(z**2)
             z = torch.clamp(z, max=self._cap) #clamp to cap
             return z

    batch_dict = ... #some batch_dict, containing tensors in keys 'input.preds' and 'groundtruth.targets'
    my_loss = MyCustomLoss(pred='input.preds', target='groundtruth.target', cap=0.4 )
    loss_val = my_loss(batch_dict)
    ...


    '''
    Advanced example - L2 regularization - in this case there are no predictions or ground truth involved, only the model weights/params
    from fuse.dl.losses import LossBase
    Class MyL2Loss(LossBase):
          '''
          Extracts prediction and data from batch_dict and performs a dummy loss calculation,
          clamping the loss value to be at most "cap"
          '''
          def __init__(self,
             params:str,
             ):
             self._params = params

          def forward(self, batch_dict:NDict) -> torch.Tensor:
             #extract pred and target from batch_dict
             params = batch_dict[self.params]
             l2_loss = sum(torch.linalg.norm(p, 2) for p in params)
             return l2_loss


    batch_dict = ... #some batch_dict, containing a key 'model.regularize_params' mapping to a list of model params to be considered in the L2 loss calculation
    loss_func = MyL2Loss(params='model.regularize_params')
    loss = loss_func(batch_dict))
    '''




    """

    @abstractmethod
    def forward(self, batch_dict: NDict) -> torch.Tensor:
        raise NotImplementedError



import time


class Timer:
    def __init__(self, txt: str = ""):
        self.txt = txt

    def __enter__(self):
        if len(self.txt) > 0:
            print("starting Timer for ", self.txt)
        self.start = time.time()

    def __exit__(self, *args):
        self.end = time.time()
        print(
            f"{self.txt} took {self.end-self.start} seconds."
        )  # todo: use minutes, hours, etc. depending on the value


if __name__ == "__main__":

    def foo(x):
        with Timer("banana"):
            for _ in range(10**5):
                x = x * 2

    foo(20)



from typing import Callable
from fuse.data.ops import get_function_call_str
from fuse.data.ops.caching_tools import get_callers_string_description, value_to_string


class HashableClass:
    _MISSING_SUPER_INIT_ERR_MSG = "Did you forget to call super().__init__() You must call it in your __init__? Also, make sure you call it BEFORE setting any attribute."

    # controls how values are converted to string, override if you want custom behavior
    # expected signature: foo(val:Any) -> str
    VALUE_TO_STRING_FUNC: Callable = value_to_string

    def __init__(self):
        """
        when init is called, a string representation of the caller(s) init args are recorded.
        This is used in get_hashable_string_representation which is used later for hashing in caching related tools (for example, SamplesCacher)
        """
        # the following is used to extract callers args, for __init__ calls up the stack of classes inheirting from OpBase
        # this way it can happen in the base class and then anyone creating new Ops will typically only need to add
        # super().__init__ in their __init__ implementation
        self._stored_init_str_representation = get_callers_string_description(
            max_look_up=4,
            expected_class=HashableClass,
            expected_function_name="__init__",
            value_to_string_func=HashableClass.VALUE_TO_STRING_FUNC,
            ignore_first_frames=3,
        )

    def __setattr__(self, name, value):
        """
        Verifies that super().__init__() is called before setting any attribute
        """
        storage_name = "_stored_init_str_representation"
        if name != storage_name and not hasattr(self, storage_name):
            raise Exception(HashableClass._MISSING_SUPER_INIT_ERR_MSG)
        super().__setattr__(name, value)

    def get_hashable_string_representation(self) -> str:
        """
        A string representation of this operation, which will be used for hashing.
        It includes recorded (string) data describing the args that were used in __init__()
        you can override/extend it in the rare cases that it's needed.

        Note - not using __str__ or __repr__ by design, to avoid cases that developers override their Ops without being aware of the effect on caching

        example:

        class OpSomethingNew(OpBase):
            def __init__(self):
                super().__init__()
            def __str__(self):
                ans = super().__str__(self)
                ans += 'whatever you want to add"

        """

        if not hasattr(self, "_stored_init_str_representation"):
            raise Exception(HashableClass._MISSING_SUPER_INIT_ERR_MSG)
        call_repr = get_function_call_str(
            self.__call__,
        )

        return f"init_{self._stored_init_str_representation}@call_{call_repr}"



from typing import Tuple, Union
import numpy as np
import torch
import skimage
from fuse.utils.ndict import NDict

from fuse.data.ops.op_base import OpBase

from fuseimg.utils.typing.key_types_imaging import DataTypeImaging
from fuseimg.data.ops.ops_common_imaging import OpApplyTypesImaging


class OpClip(OpBase):
    """
    Clip values - support both torch tensor and numpy array
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(
        self,
        sample_dict: NDict,
        key: str,
        clip=(0.0, 1.0),
    ):
        """
        Clip  values
        :param key: key to an image in sample_dict: either torh tensor or numpy array and any dimension
        :param clip: values for clipping from both sides
        """

        img = sample_dict[key]

        processed_img = self.clip(img, clip)

        sample_dict[key] = processed_img
        return sample_dict

    @staticmethod
    def clip(
        img: Union[np.ndarray, torch.Tensor], clip: Tuple[float, float] = (0.0, 1.0)
    ) -> Union[np.ndarray, torch.Tensor]:
        if isinstance(img, np.ndarray):
            processed_img = np.clip(img, clip[0], clip[1])
        elif isinstance(img, torch.Tensor):
            processed_img = torch.clamp(img, clip[0], clip[1], out=img)
        else:
            raise Exception(f"Error: unexpected type {type(img)}")
        return processed_img


op_clip_img = OpApplyTypesImaging({DataTypeImaging.IMAGE: (OpClip(), {})})


class OpNormalizeAgainstSelf(OpBase):
    """
    normalizes a tensor into [0.0, 1.0] using its own statistics (NOT against a dataset)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, sample_dict: NDict, key: str):
        img = sample_dict[key]
        img = skimage.img_as_float(img)
        sample_dict[key] = img

        return sample_dict


op_normalize_against_self_img = OpApplyTypesImaging({DataTypeImaging.IMAGE: (OpNormalizeAgainstSelf(), {})})


class OpToIntImageSpace(OpBase):
    """
    normalizes a tensor into [0, 255] int gray-scale using its own statistics (NOT against a dataset)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(
        self,
        sample_dict: NDict,
        key: str,
    ):
        img = sample_dict[key]
        img -= img.min()
        img /= img.max()
        img *= 255.0
        img = img.astype(np.uint8).copy()
        # img = img.transpose((1, 2, 0))
        sample_dict[key] = img
        return sample_dict


op_to_int_image_space_img = OpApplyTypesImaging({DataTypeImaging.IMAGE: (OpToIntImageSpace(), {})})


class OpToRange(OpBase):
    """
    linearly project from a range to a different range
    """

    def __call__(
        self,
        sample_dict: NDict,
        key: str,
        from_range: Tuple[float, float],
        to_range: Tuple[float, float],
    ):

        from_range_start = from_range[0]
        from_range_end = from_range[1]
        to_range_start = to_range[0]
        to_range_end = to_range[1]

        img = sample_dict[key]

        # shift to start at 0
        img -= from_range_start

        # scale to be in desired range
        img *= (to_range_end - to_range_start) / (from_range_end - from_range_start)
        # shift to start in desired start val
        img += to_range_start

        sample_dict[key] = img

        return sample_dict


op_to_range_img = OpApplyTypesImaging({DataTypeImaging.IMAGE: (OpToRange(), {})})



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
# FIXME: data_package
import multiprocessing
from fuse.utils import NDict
import unittest
import shutil
import tempfile
import os
from fuse.utils.gpu import choose_and_enable_multiple_gpus
from fuse.utils.multiprocessing.run_multiprocessed import run_in_subprocess


# os env variable CMMD_DATA_PATH is a path to the stored dataset location
# dataset should be download from https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=70230508
# download requires NBIA data retriever https://wiki.cancerimagingarchive.net/display/NBIA/Downloading+TCIA+Images
# put on the folliwing in the main folder  -
# 1. CMMD_clinicaldata_revision.csv which is a converted version of CMMD_clinicaldata_revision.xlsx
# 2. folder named CMMD which is the downloaded data folder
if "CMMD_DATA_PATH" in os.environ:
    from fuse_examples.imaging.classification.cmmd.runner import run_train, run_eval, run_infer


def run_cmmd(root: str) -> None:
    cfg = NDict(
        {
            "paths": {
                "data_dir": os.environ["CMMD_DATA_PATH"],
                "model_dir": os.path.join(root, "model_new/InceptionResnetV2_2017_test"),
                "inference_dir": os.path.join(root, "model_new/infer_dir"),
                "eval_dir": os.path.join(root, "model_new/eval_dir"),
                "cache_dir": os.path.join(root, "examples/CMMD_cache_dir"),
                "data_misc_dir": os.path.join(root, "data_misc"),
                "data_split_filename": "cmmd_split.pkl",
            },
            "run": {"running_modes": ["train", "infer", "eval"]},
            "train": {
                "target": "classification",
                "reset_cache": False,
                "num_workers": 10,
                "num_folds": 5,
                "train_folds": [0, 1, 2],
                "validation_folds": [3],
                "batch_size": 2,
                "learning_rate": 0.0001,
                "weight_decay": 0,
                "resume_checkpoint_filename": None,
                "trainer": {"accelerator": "gpu", "devices": 1, "num_epochs": 2, "ckpt_path": None},
            },
            "infer": {
                "infer_filename": "validation_set_infer.gz",
                "checkpoint": "best_epoch.ckpt",
                "infer_folds": [4],
                "target": "classification",
                "num_workers": 10,
            },
        }
    )
    print(cfg)
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    choose_and_enable_multiple_gpus(cfg["train.trainer.devices"], force_gpus=force_gpus)

    run_train(cfg["paths"], cfg["train"])
    run_infer(cfg["train"], cfg["paths"], cfg["infer"])
    results = run_eval(cfg["paths"], cfg["infer"])

    assert "metrics.auc" in results


@unittest.skipIf("CMMD_DATA_PATH" not in os.environ, "define environment variable 'CMMD_DATA_PATH' to run this test")
class ClassificationMGCmmdTestCase(unittest.TestCase):
    def setUp(self):
        self.root = tempfile.mkdtemp()

    def test_runner(self):
        run_in_subprocess(run_cmmd, self.root)

    def tearDown(self):
        # Delete temporary directories
        shutil.rmtree(self.root)


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, Tuple, Sequence, Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from fuse.dl.models.heads.common import ClassifierFCN
from fuse.utils.ndict import NDict


class HeadDenseSegmentation(nn.Module):
    def __init__(
        self,
        head_name: str = "head_0",
        conv_inputs: Sequence[Tuple[str, int]] = None,
        num_classes: int = 2,
        post_concat_inputs: Optional[Sequence[Tuple[str, int]]] = None,
        maxpool_kernel: Optional[Union[Tuple[int, int], int]] = None,
        layers_description: Sequence[int] = (256,),
        dropout_rate: float = 0.1,
        shared_classifier_head: Optional[torch.nn.Module] = None,
    ) -> None:
        """
        Dense segmentation head - predicts class scores for each location of input feature map ("conv_input").

        Output of a forward pass:
        'model.logits.head_name' and 'outputs.head_name', both in shape [batch_size, num_classes, input_height, input_width]

        :param head_name:                   batch_dict key
        :param conv_inputs:                 List of feature map inputs - tuples of (batch_dict key, channel depth)
                                            If multiple inputs are used, they are concatenated on the channel axis
                example: conv_inputs=(('model.backbone_features', 384),),
        :param num_classes:                 Number of output classes (per feature map location)
        :param post_concat_inputs:          Additional vector (one dimensional) inputs, concatenated just before the classifier module
        :param maxpool_kernel:              Kernel size for an optional preliminary max pooling step, to reduce feature maps size
        :param layers_description:          Layers description for the classifier module - sequence of hidden layers sizes
        :param dropout_rate:                Dropout rate for classifier module layers
        :param shared_classifier_head:      Optional reference for external torch.nn.Module classifier
        """
        super().__init__()

        self.head_name = head_name
        assert conv_inputs is not None, "conv_inputs must be provided"
        self.conv_inputs = conv_inputs
        self.maxpool_kernel = maxpool_kernel

        feature_depth = sum([conv_input[1] for conv_input in self.conv_inputs])
        if post_concat_inputs is not None:
            feature_depth += sum([post_concat_input[1] for post_concat_input in post_concat_inputs])

        if shared_classifier_head is not None:
            self.classifier_head_module = shared_classifier_head
        else:
            self.classifier_head_module = ClassifierFCN(
                in_ch=feature_depth,
                num_classes=num_classes,
                layers_description=layers_description,
                dropout_rate=dropout_rate,
            )

    def forward(self, batch_dict: NDict) -> Dict:
        conv_input = torch.cat([batch_dict[conv_input[0]] for conv_input in self.conv_inputs])
        if self.maxpool_kernel is not None:
            conv_input = F.max_pool2d(conv_input, kernel_size=self.maxpool_kernel)

        logits = self.classifier_head_module(conv_input)
        score_map = F.softmax(logits, dim=1)  # --> score_map.shape = [batch_size, 2, height, width]

        batch_dict["model.logits." + self.head_name] = logits
        batch_dict["model.output." + self.head_name] = score_map

        return batch_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import unittest

from fuse.utils.rand.seed import Seed

# from fuse.utils.file_io.file_io import SAFE_save_hdf5, load_hdf5
from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.data import get_sample_id, create_initial_sample
import numpy as np
import tempfile
import os
from fuse.data.ops.op_base import OpBase
from typing import List, Union, Optional
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.utils.ndict import NDict


class OpFakeLoad(OpBase):
    def __init__(self):
        super().__init__()

    def __call__(self, sample_dict: NDict, **kwargs) -> Union[None, dict, List[dict]]:
        sid = get_sample_id(sample_dict)
        if "case_1" == sid:
            sample_dict.merge(_generate_sample_1())
        elif "case_2" == sid:
            sample_dict.merge(_generate_sample_2())
        elif "case_3" == sid:
            return None
        elif "case_4" == sid:
            sample_1 = create_initial_sample("case_4", "case_4_subcase_1")
            sample_1.merge(_generate_sample_1(41))

            sample_2 = create_initial_sample("case_4", "case_4_subcase_2")
            sample_2.merge(_generate_sample_2(42))

            return [sample_1, sample_2]
        # elif 'case_4_subcase_1' == sid:
        #     sample_dict.merge(_generate_sample_1(41))
        # elif 'case_4_subcase_2' == sid:
        #     sample_dict.merge(_generate_sample_2(42))
        else:
            raise Exception(f"unfamiliar sample_id: {sid}")
        return sample_dict


class OpPrintContents(OpBase):
    def __init__(self):
        super().__init__()

    def __call__(self, sample_dict: NDict, **kwargs) -> Union[None, dict, List[dict]]:
        sid = get_sample_id(sample_dict)
        print(f"sid={sid}")
        for k in sample_dict.keypaths():
            print(k)
        print("-------------------------\n")

        return sample_dict


class TestDatasetDefault(unittest.TestCase):
    """
    Test sample caching
    """

    def setUp(self):
        pass

    def test_cache_samples_with_sample_morphing(self):
        tmpdir = tempfile.mkdtemp()
        cache_dirs = [
            os.path.join(tmpdir, "cache_a"),
            os.path.join(tmpdir, "cache_b"),
        ]

        static_pipeline_desc = [
            (OpFakeLoad(), {}),
        ]

        dynamic_pipeline_desc = [
            (OpPrintContents(), {}),
        ]

        static_pl = PipelineDefault(
            "static_pipeline",
            static_pipeline_desc,
        )
        dynamic_pl = PipelineDefault(
            "dynamic_pipeline",
            dynamic_pipeline_desc,
        )

        orig_sample_ids = ["case_1", "case_2", "case_3", "case_4"]
        ################ cached + sample morphing
        cacher = SamplesCacher("dataset_test_cache", static_pl, cache_dirs, restart_cache=True)

        ds_cached = DatasetDefault(
            orig_sample_ids,
            static_pl,
            dynamic_pipeline=dynamic_pl,
            cacher=cacher,
        )

        ds_cached.create(num_workers=0)
        cached_final_sample_ids = ds_cached.get_all_sample_ids()

        ############### not cached + sample morphing

        ds_not_cached = DatasetDefault(
            orig_sample_ids,
            static_pl,
            dynamic_pipeline=dynamic_pl,
            cacher=None,
            allow_uncached_sample_morphing=True,
        )
        ds_not_cached.create(num_workers=0)
        not_cached_final_sample_ids = ds_not_cached.get_all_sample_ids()

        self.assertEqual(
            sorted(cached_final_sample_ids),
            sorted(not_cached_final_sample_ids),
        )

        sample_from_cached = ds_cached[3]
        sample_from_not_cached = ds_not_cached[3]

        self.assertEqual(
            sample_from_cached["data"]["cc"]["img"].sum(), sample_from_not_cached["data"]["cc"]["img"].sum()
        )

        self.assertEqual(sample_from_cached["data"]["cc"]["img"].sum(), 49948.825007353706)
        banana = 123

    def test_cache_samples_no_sample_morphing(self):
        tmpdir = tempfile.mkdtemp()
        cache_dirs = [
            os.path.join(tmpdir, "cache_a"),
            os.path.join(tmpdir, "cache_b"),
        ]

        static_pipeline_desc = [
            (OpFakeLoad(), {}),
        ]

        dynamic_pipeline_desc = [
            (OpPrintContents(), {}),
        ]

        static_pl = PipelineDefault(
            "static_pipeline",
            static_pipeline_desc,
        )
        dynamic_pl = PipelineDefault(
            "dynamic_pipeline",
            dynamic_pipeline_desc,
        )

        orig_sample_ids = ["case_1", "case_2"]
        ################ cached + no sample morphing
        cacher = SamplesCacher("dataset_test_cache", static_pl, cache_dirs, restart_cache=True)

        ds_cached = DatasetDefault(
            orig_sample_ids,
            static_pl,
            dynamic_pipeline=dynamic_pl,
            cacher=cacher,
        )

        ds_cached.create(num_workers=0)
        cached_final_sample_ids = ds_cached.get_all_sample_ids()

        ############### not cached + no sample morphing

        ds_not_cached = DatasetDefault(
            orig_sample_ids,
            static_pl,
            dynamic_pipeline=dynamic_pl,
            cacher=None,
            ###allow_uncached_sample_morphing=False,
        )
        ds_not_cached.create(num_workers=0)
        not_cached_final_sample_ids = ds_not_cached.get_all_sample_ids()

        self.assertEqual(
            sorted(cached_final_sample_ids),
            sorted(not_cached_final_sample_ids),
        )

        sample_from_cached = ds_cached[1]
        sample_from_not_cached = ds_not_cached[1]

        self.assertEqual(
            sample_from_cached["data"]["cc"]["img"].sum(), sample_from_not_cached["data"]["cc"]["img"].sum()
        )

        self.assertEqual(sample_from_cached["data"]["cc"]["img"].sum(), 50012.88698394645)
        banana = 123

    def tearDown(self):
        pass


def _generate_sample_1(seed=1337):
    Seed.set_seed(seed)
    sample = dict(
        data=dict(
            cc=dict(
                img=np.random.rand(30, 200, 200),
                seg=np.random.randint(0, 16, size=(30, 200, 200)),
                dicom_tags=[10, 13, 40, "banana"],
            ),
            mlo=dict(
                img=np.random.rand(30, 200, 200),
                seg=np.random.randint(0, 16, size=(40, 100, 164)),
                dicom_tags=[100, 130, 400, "banana123"],
            ),
            gt_labels_style_1=[1, 3, 100, 12],
            gt_labels_style_2=np.array([3, 4, 10, 12]),
            clinical_info_input=np.random.rand(1000),
        )
    )
    return sample


def _generate_sample_2(seed=1234):
    Seed.set_seed(seed)
    sample = dict(
        data=dict(
            cc=dict(
                img=np.random.rand(10, 100, 100),
                seg=np.random.randint(0, 16, size=(10, 100, 10)),
                dicom_tags=[20, 23, 60, "bananaphone"],
            ),
            mlo=dict(
                img=np.random.rand(30, 200, 200),
                seg=np.random.randint(0, 16, size=(40, 100, 164)),
                dicom_tags=[12, 13, 40, "porcupine123"],
            ),
            gt_labels_style_1=[5, 2, 13, 16],
            gt_labels_style_2=np.array([8, 14, 11, 1]),
            clinical_info_input=np.random.rand(90),
        )
    )
    return sample


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from typing import Any, Callable, Dict, Optional, Sequence, Union, Hashable

import numpy as np
import pandas as pd

from fuse.eval.metrics.metrics_common import MetricBase, MetricWithCollectorBase
from fuse.eval.metrics.libs.model_comparison import ModelComparison


class PairedBootstrap(MetricWithCollectorBase):
    """
    Verify your hypothesis about the relation between models using bootstrap approach
    The specific hypothesis can be specified in compare_method which by default checks superiority.
    Typically, the compare_method will compute p_value:
    the probability of obtaining test results at least as extreme as the results actually observed,
    under the assumption that the null hypothesis is correct.
    The format of the result dictionary returned by eval() method may varied and depends in comapre_method used
    """

    def __init__(
        self,
        metric_test: MetricBase,
        metric_reference: MetricBase,
        stratum: str,
        metric_keys_to_compare: Optional[Sequence[str]] = None,
        compare_method: Callable = ModelComparison.bootstrap_margin_superiority,
        num_of_bootstraps: int = 10000,
        rnd_seed: int = 1234,
        margin: float = 0.0,
        **super_kwargs
    ) -> None:
        """
        :param metric_test: compare the results of metric_test to results of metric_reference
        :param metric_reference: compare the results of metric_test to results of metric_reference
        :param stratum: if the resampling should be done by strata, specify the key in batch_dict to collect data from
        :param num_of_bootstraps: number of bootstrap iterations
        :param rnd_seed: seed for random number generator.
        :param margin: the difference required to be considered significant by the compare_method
        :param metrics_keys_to_compare: compare the following metrics values: keys in the dictionary returned by metric_test and metric_reference.
                                        Set to None if those metrics returns single value.
        :param compare_method: callback that defines the hypothesis and compares the results.
                               Required prototype: func(name: str, test_values: Sequence[float], reference_values: Sequence[float]) -> Union[Dict[str, float], float]
                               See example: ModelComparison.bootstrap_margin_superiority , ModelComparison.bootstrap_margin_non_inferiority , ModelComparison.bootstrap_margin_equality
        """

        super().__init__(stratum=stratum, **super_kwargs)

        # store arguments
        self._metric_test = metric_test
        self._metric_reference = metric_reference
        self._num_of_bootstraps = num_of_bootstraps
        self._rnd_seed = rnd_seed
        self._metric_keys_to_compare = metric_keys_to_compare
        self._compare_method = compare_method
        self._margin = margin

    def collect(self, batch: Dict) -> None:
        "See super class"
        self._metric_test.collect(batch)
        self._metric_reference.collect(batch)
        return super().collect(batch)

    def set(self, data: Union[Dict, Sequence[Dict], pd.DataFrame]) -> None:
        "See super class"
        self._metric_test.set(data)
        self._metric_reference.set(data)
        return super().set(data)

    def reset(self) -> None:
        "See super class"
        self._metric_test.reset()
        self._metric_reference.reset()
        return super().reset()

    def eval(self, results: Dict[str, Any] = None, ids: Sequence[Hashable] = None) -> Dict[str, Any]:
        """
        See super class
        :return: the compared results dictionary created by compare_method() specified in constructor
        """
        # prepare data
        data = self._collector.get(ids)
        if ids is None:
            ids = self._collector.get_ids()
        if not ids:
            raise Exception(
                "Error: paired bootstrap is supported only when a unique identifier is specified. Add key 'id' to your data"
            )
        ids = np.array(ids)

        rnd = np.random.RandomState(self._rnd_seed)
        stratum_id = np.array(data["stratum"]) if "stratum" in data else np.ones(len(data["id"]))
        unique_strata = np.unique(stratum_id)

        # initialize results
        if self._metric_keys_to_compare is None:
            bootstrap_results_test = np.empty(self._num_of_bootstraps)
            bootstrap_results_reference = np.empty(self._num_of_bootstraps)
        else:
            bootstrap_results_test = {key: np.empty(self._num_of_bootstraps) for key in self._metric_keys_to_compare}
            bootstrap_results_reference = {
                key: np.empty(self._num_of_bootstraps) for key in self._metric_keys_to_compare
            }

        # aggregate bootstrap results
        for bs_index in range(self._num_of_bootstraps):
            sampled_ids = ids.copy()
            for stratum in unique_strata:
                stratum_filter = stratum_id == stratum
                n_stratum = sum(stratum_filter)
                random_sample = rnd.randint(0, n_stratum, size=n_stratum)
                sampled_ids[stratum_filter] = ids[stratum_filter][random_sample]
            result_test = self._metric_test.eval(results, sampled_ids)
            result_reference = self._metric_reference.eval(results, sampled_ids)

            # single-value metric case
            if self._metric_keys_to_compare is None:
                assert not isinstance(
                    result_test, Dict
                ), "Error: metric_test returned a dictionary of results. Specify the relevant keys to compare in metric_keys_to_compare."
                assert not isinstance(
                    result_reference, Dict
                ), "Error: metric_reference returned a dictionary of results. Specify the relevant keys to compare in metric_keys_to_compare."
                bootstrap_results_test[bs_index] = result_test
                bootstrap_results_reference[bs_index] = result_reference

            # multi-values metric case
            if self._metric_keys_to_compare is not None:
                assert isinstance(
                    result_test, Dict
                ), "Error: metric_test returned a single value. Set metric_keys_to_compare to None."
                assert isinstance(
                    result_reference, Dict
                ), "Error: metric_reference returned a single value. Set metric_keys_to_compare to None."
                for key in self._metric_keys_to_compare:
                    bootstrap_results_test[key][bs_index] = result_test[key]
                    bootstrap_results_reference[key][bs_index] = result_reference[key]

        # compare
        if self._metric_keys_to_compare is None:
            return self._compare_method(None, bootstrap_results_test, bootstrap_results_reference, margin=self._margin)

        metric_results = {}
        for key in bootstrap_results_test:
            metric_results[key] = self._compare_method(
                key, bootstrap_results_test[key], bootstrap_results_reference[key], margin=self._margin
            )
        return metric_results



import numpy as np
from typing import List, Tuple


def get_chunks_ranges(total: int, chunk_size: int) -> List[Tuple[int, int]]:
    """
    Creates "chunks" of work, useful when creating worker functions for run_multiprocessed
     where each workers gets a range ("chunk") to work on.

    Returns a list of tuples [start_index, end_index)

    for example: get_chunks_ranges(10, 3) would return [(0,3), (3,6), (6,9), (9,10)]
    """

    if chunk_size >= total:
        return [(0, total)]

    steps = np.arange(0, total, chunk_size)
    ans = list(zip(steps[:-1], steps[1:]))
    if ans[-1][-1] < total:
        ans.append((ans[-1][-1], total))
    return ans



import unittest

from typing import Union, List
from fuse.utils.ndict import NDict

from fuse.data.ops.op_base import OpBase, op_call
from fuse.data.key_types import DataTypeBasic
from fuse.data import create_initial_sample
from fuse.data.key_types_for_testing import DataTypeForTesting, type_detector_for_testing


class TestOpBase(unittest.TestCase):
    def test_for_type_detector(self):
        td = type_detector_for_testing
        sample = create_initial_sample("dummy")

        self.assertEqual(td.get_type(sample, "data.cc.img_for_testing"), DataTypeForTesting.IMAGE_FOR_TESTING)
        self.assertEqual(td.get_type(sample, "data.cc_img_for_testing"), DataTypeForTesting.IMAGE_FOR_TESTING)
        self.assertEqual(td.get_type(sample, "data.img_seg_for_testing"), DataTypeForTesting.SEG_FOR_TESTING)
        self.assertEqual(td.get_type(sample, "data.imgseg_for_testing"), DataTypeForTesting.SEG_FOR_TESTING)
        self.assertEqual(td.get_type(sample, "data"), DataTypeBasic.UNKNOWN)
        self.assertEqual(td.get_type(sample, "bbox_for_testing"), DataTypeForTesting.BBOX_FOR_TESTING)
        self.assertEqual(td.get_type(sample, "a.bbox_for_testing"), DataTypeForTesting.BBOX_FOR_TESTING)

    def test_op_base(self):
        class OpImp(OpBase):
            def __call__(self, sample_dict: NDict, **kwargs) -> Union[None, dict, List[dict]]:
                sample_dict["data.cc.seg_for_testing"] = 5
                return sample_dict

        op = OpImp()
        sample_dict = {}
        sample_dict = op_call(op, sample_dict, "id")
        self.assertTrue("data.cc.seg_for_testing" in sample_dict)
        self.assertTrue(sample_dict["data.cc.seg_for_testing"] == 5)
        self.assertTrue(
            type_detector_for_testing.get_type(sample_dict, "data.cc.seg_for_testing")
            == DataTypeForTesting.SEG_FOR_TESTING
        )
        type_detector_for_testing.verify_type(
            sample_dict, "data.cc.seg_for_testing", [DataTypeForTesting.SEG_FOR_TESTING]
        )
        self.assertRaises(
            ValueError,
            type_detector_for_testing.verify_type,
            sample_dict,
            "data.cc.seg_for_testing",
            [DataTypeForTesting.IMAGE_FOR_TESTING],
        )


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import gzip
import os
import threading


class AtomicFileWriter:
    """Writes a file to filename only on successful completion"""

    def __init__(self, filename: str):
        self.filename = filename
        self.temp_filename = f"{filename}_{os.getpid()}_{threading.get_ident() }.tmp"

    def __enter__(self):
        if self.filename.endswith(".gz"):
            self.filehandle = gzip.open(self.temp_filename, "wb")
        else:
            self.filehandle = open(self.temp_filename, "wb")
        return self.filehandle

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is None:
            self.filehandle.close()
            os.replace(self.temp_filename, self.filename)
        else:
            try:
                self.filehandle.close()
            finally:
                os.unlink(self.temp_filename)



from collections import OrderedDict
import pathlib
from fuse.utils.utils_logger import fuse_logger_start
import os
import yaml

import pandas as pd
from fuse.dl.models import ModelMultiHead
from fuse.dl.models.backbones.backbone_resnet_3d import BackboneResnet3D
from fuse.dl.models.heads.heads_3D import Head3DClassifier
from fuseimg.datasets.knight import KNIGHT
import torch.nn.functional as F
import torch.nn as nn
from torch.utils.data.dataloader import DataLoader
from fuse.data.utils.collates import CollateDefault
from fuse.data.utils.samplers import BatchSamplerDefault

from fuse.eval.metrics.classification.metrics_classification_common import MetricAUCROC, MetricAccuracy, MetricConfusion
from fuse.eval.metrics.classification.metrics_thresholding_common import MetricApplyThresholds
import torch.optim as optim
import fuse.utils.gpu as GPU
from fuse.utils.rand.seed import Seed
import logging
import time
import copy
from fuse.dl.losses.loss_default import LossDefault
from fuse.dl.lightning.pl_module import LightningModuleDefault
import pytorch_lightning as pl

## Parameters:
##############################################################################
# Data sources to use in model. Set {'imaging': True, 'clinical': False} for imaging only setting,
# and vice versa, or set both to True to use both.
# allocate gpus
# uncomment if you want to use specific gpus instead of automatically looking for free ones


def make_model(use_data: dict, num_classes: int, imaging_dropout: float, fused_dropout: float):
    if use_data["imaging"]:
        backbone = BackboneResnet3D(in_channels=1)
        conv_inputs = [("model.backbone_features", 512)]
    else:
        backbone = nn.Identity()
        conv_inputs = None
    if use_data["clinical"]:
        append_features = [("data.input.clinical.all", 11)]
    else:
        append_features = None

    model = ModelMultiHead(
        conv_inputs=(("data.input.img", 1),),
        backbone=backbone,
        heads=[
            Head3DClassifier(
                head_name="head_0",
                conv_inputs=conv_inputs,
                dropout_rate=imaging_dropout,
                num_classes=num_classes,
                append_features=append_features,
                append_layers_description=(256, 128),
                fused_dropout_rate=fused_dropout,
            ),
        ],
    )
    return model


def main(cfg_path):
    # read config params
    cfg = yaml.safe_load(open(cfg_path))
    task_num = cfg["task_num"]
    num_classes = cfg[task_num]["num_classes"]
    target_name = cfg[task_num]["target_name"]
    # read train/val splits file. for convenience, we use the one
    # auto-generated by the nnU-Net framework for the KiTS21 data
    dir_path = pathlib.Path(__file__).parent.resolve()
    splits = pd.read_pickle(os.path.join(dir_path, "splits_final.pkl"))
    # For this example, we use split 0 out of the 5 available cross validation splits
    split = splits[0]
    if cfg["testing"]:
        split["train"] = split["train"][:5]
        split["val"] = split["val"][:5]
    # read environment variables for data, cache and results locations
    data_path = os.environ["KNIGHT_DATA"]
    cache_path = os.path.join(os.environ["KNIGHT_CACHE"], str(cfg["experiment_num"]))
    results_path = os.environ["KNIGHT_RESULTS"]

    ## Basic settings:
    ##############################################################################
    # create model results dir:
    # we use a time stamp in model directory name, to prevent re-writing
    timestr = time.strftime("%Y%m%d-%H%M%S")
    model_dir = os.path.join(results_path, timestr)
    if not os.path.isdir(model_dir):
        os.makedirs(model_dir)

    # start logger
    fuse_logger_start(output_path=model_dir, console_verbose_level=logging.INFO)
    print("Done")

    # set constant seed for reproducibility.
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"  # required for pytorch deterministic mode
    rand_gen = Seed.set_seed(1234, deterministic_mode=True)

    # select gpus
    GPU.choose_and_enable_multiple_gpus(cfg["num_gpus"], force_gpus=None)

    ## Model definition
    ##############################################################################

    model = make_model(cfg["use_data"], num_classes, cfg["imaging_dropout"], cfg["fused_dropout"])

    ## FuseMedML dataset preparation
    ##############################################################################

    train_ds, valid_ds = KNIGHT.dataset(
        data_path=data_path,
        cache_dir=cache_path,
        split=split,
        reset_cache=False,
        resize_to=cfg["resize_to"],
    )

    ## Create dataloader

    sampler = BatchSamplerDefault(
        dataset=train_ds,
        balanced_class_name=target_name,
        num_balanced_classes=num_classes,
        batch_size=cfg["batch_size"],
        balanced_class_weights=[1.0 / num_classes] * num_classes if task_num == "task_2" else None,
        mode="approx",
    )

    train_dl = DataLoader(
        dataset=train_ds,
        shuffle=False,
        drop_last=False,
        batch_sampler=sampler,
        collate_fn=CollateDefault(),
        num_workers=8,
    )

    valid_dl = DataLoader(
        dataset=valid_ds,
        shuffle=False,
        drop_last=False,
        batch_sampler=None,
        batch_size=cfg["batch_size"],
        num_workers=8,
        collate_fn=CollateDefault(),
        generator=rand_gen,
    )

    # Loss definition:
    ##############################################################################
    losses = {
        "cls_loss": LossDefault(pred="model.logits.head_0", target=target_name, callable=F.cross_entropy, weight=1.0),
    }

    # Metrics definition:
    ##############################################################################
    train_metrics = OrderedDict(
        [
            ("op", MetricApplyThresholds(pred="model.output.head_0")),  # will apply argmax
            ("auc", MetricAUCROC(pred="model.output.head_0", target=target_name)),
            ("accuracy", MetricAccuracy(pred="results:metrics.op.cls_pred", target=target_name)),
            (
                "sensitivity",
                MetricConfusion(pred="results:metrics.op.cls_pred", target=target_name, metrics=("sensitivity",)),
            ),
        ]
    )
    val_metrics = copy.deepcopy(train_metrics)  # use the same metrics in validation as well

    best_epoch_source = dict(
        monitor=cfg[task_num]["target_metric"],  # can be any key from losses or metrics dictionaries
        mode="max",  # can be either min/max
    )

    # Optimizer definition:
    ##############################################################################
    optimizer = optim.Adam(model.parameters(), lr=cfg["learning_rate"], weight_decay=0.001)

    # Scheduler definition:
    ##############################################################################
    lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer)
    lr_sch_config = dict(scheduler=lr_scheduler, monitor="validation.losses.total_loss")
    optimizers_and_lr_schs = dict(optimizer=optimizer, lr_scheduler=lr_sch_config)

    ## Training
    ##############################################################################

    # create instance of PL module - FuseMedML generic version
    pl_module = LightningModuleDefault(
        model_dir=model_dir,
        model=model,
        losses=losses,
        train_metrics=train_metrics,
        validation_metrics=val_metrics,
        best_epoch_source=best_epoch_source,
        optimizers_and_lr_schs=optimizers_and_lr_schs,
    )
    # create lightining trainer.
    pl_trainer = pl.Trainer(
        default_root_dir=model_dir,
        max_epochs=cfg["num_epochs"],
        accelerator="gpu",
        devices=cfg["num_gpus"],
        strategy=None,
        auto_select_gpus=True,
        num_sanity_val_steps=-1,
    )

    # train
    pl_trainer.fit(pl_module, train_dl, valid_dl, ckpt_path=None)


if __name__ == "__main__":
    config_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "config.yaml")
    main(config_path)



# version
from fuse.version import __version__

# import shortcuts
from fuse.data.utils.sample import get_sample_id, set_sample_id, get_sample_id_key
from fuse.data.utils.sample import (
    create_initial_sample,
    get_initial_sample_id,
    get_initial_sample_id_key,
    get_specific_sample_from_potentially_morphed,
)
from fuse.data.ops.op_base import OpBase  # DataTypeForTesting,
from fuse.data.ops.ops_common import OpApplyPatterns, OpLambda, OpFunc, OpRepeat, OpKeepKeypaths
from fuse.data.ops.ops_aug_common import OpRandApply, OpSample, OpSampleAndRepeat
from fuse.data.ops.ops_read import OpReadDataframe
from fuse.data.ops.ops_cast import OpToTensor, OpToNumpy
from fuse.data.utils.collates import CollateDefault
from fuse.data.utils.export import ExportDataset
from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.data.datasets.dataset_default import DatasetBase, DatasetDefault



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Callable, Optional

import torch

from fuse.dl.losses.loss_base import LossBase
from fuse.utils.ndict import NDict


class LossDefault(LossBase):
    """
    Default Fuse loss function

    Basic Usage Example:
    '''
    from fuse.dl.losses.loss_default import LossDefault

    my_loss_func = LossDefault(
        pred='model.preds',
        target='data.groundtruth.targets',
        callable=torch.nn.functional.cross_entropy,
        )

    batch_dict = # batch_dict with pred and target
    loss = my_loss_func(batch_dict)
    '''


    Custom Preprocessing Example:
    Sometimes you want custom preprocessing of the batch_dict - for example in the following scenario:

    A multi-head / multi-task model, in which you have ground truth labels only for a subset of the samples.
    In such case, you may use the optional "preprocess_func" to filter out the samples that you don't have labels for both tasks.

    Example code:
    '''
    def my_preprocess_func(batch_dict: NDict) -> NDict:
        # filter out samples that don't have labels for both tasks
        keep_indices = #... calculate which indices to keep
        batch_dict = batch_dict.indices(keep_indices) #this will keep only a subset of elements per each key in the dictionary (where applicable)
        return batch_dict

    my_loss_func = LossDefault(
        pred='model.preds',
        target='data.groundtruth.targets',
        callable=torch.nn.functional.cross_entropy,
        preprocess_func=my_preprocess_func,
        )
    '''

    """

    def __init__(
        self,
        *,  # prevent positional args
        pred: str = None,
        target: str = None,
        callable: Callable = None,
        weight: Optional[float] = None,
        preprocess_func: Optional[Callable] = None,
    ) -> None:
        """
        This class wraps a PyTorch loss function with a Fuse api.
        Args:
        :param pred:               batch_dict key for prediction (e.g., network output)
        :param target:             batch_dict key for target (e.g., ground truth label)
        :param callable:           PyTorch loss function handle (e.g., torch.nn.functional.cross_entropy)
        :param weight:             scalar loss multiplier
        :param preprocess_func:             function that filters batch_dict/ The function gets an input batch_dict and returns filtered batch_dict
            the expected function signature is:
                foo(batch_dict: NDict) -> NDict:
        """
        super().__init__()
        self.pred = pred
        self.target = target
        self.callable = callable
        self.weight = weight
        self.preprocess_func = preprocess_func

    def forward(self, batch_dict: NDict) -> torch.Tensor:
        # preprocess batch_dict if required
        if self.preprocess_func is not None:
            batch_dict = self.preprocess_func(batch_dict)
        preds = batch_dict[self.pred]
        targets = batch_dict[self.target]

        loss_obj = self.callable(preds, targets)

        if self.weight is not None:
            loss_obj *= self.weight

        return loss_obj



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Sequence, Dict, Tuple

import torch

from fuse.utils.ndict import NDict


class ModelMultiHead(torch.nn.Module):
    """
    Default Fuse model - convolutional neural network with multiple heads
    """

    def __init__(
        self,
        backbone: torch.nn.Module,
        heads: Sequence[torch.nn.Module],
        conv_inputs: Tuple[Tuple[str, int], ...] = None,
        backbone_args: Tuple[Tuple[str, int], ...] = None,
    ) -> None:
        """
        Default Fuse model - convolutional neural network with multiple heads
        :param conv_inputs:     batch_dict name for convolutional backbone model input and its number of input channels. Unused if None. Kept for backward compatibility
        :param backbone_args:   batch_dict name for generic backbone model input and its number of input channels. Unused if None
        :param backbone:        PyTorch backbone module - a convolutional (in which case conv_inputs must be supplied) or some other (in which case backbone_args must be supplied) neural network
        :param heads:           Sequence of head modules
        """
        super().__init__()
        if (conv_inputs is not None) and (backbone_args is not None):
            raise Exception(
                "Both conv_inputs and backbone_args are set. Only one may be set (conv_inputs soon to be deprecated)"
            )
        if (conv_inputs is None) and (backbone_args is None):
            raise Exception(
                "Neither conv_inputs nor backbone_args are None. One must be set (conv_inputs soon to be deprecated)"
            )

        self.conv_inputs = conv_inputs
        self.backbone_args = backbone_args
        self.backbone = backbone
        self.add_module("backbone", self.backbone)
        self.heads = torch.nn.ModuleList(heads)
        self.add_module("heads", self.heads)

    def forward(self, batch_dict: NDict) -> Dict:
        if self.conv_inputs is not None:
            conv_input = torch.cat([batch_dict[conv_input[0]] for conv_input in self.conv_inputs], 1)
            backbone_features = self.backbone.forward(conv_input)
        else:
            backbone_args = [batch_dict[inp[0]] for inp in self.backbone_args]
            backbone_features = self.backbone.forward(*backbone_args)

        batch_dict["model.backbone_features"] = backbone_features

        for head in self.heads:
            batch_dict = head.forward(batch_dict)

        return batch_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from typing import Dict, Optional, Sequence, Tuple, Union

import pandas as pd
import numpy as np

from sklearn import metrics
import sklearn

import matplotlib.pyplot as plt


class MetricsLibClass:
    @staticmethod
    def auc_roc(
        pred: Sequence[Union[np.ndarray, float]],
        target: Sequence[Union[np.ndarray, int]],
        sample_weight: Optional[Sequence[Union[np.ndarray, float]]] = None,
        pos_class_index: int = -1,
        max_fpr: Optional[float] = None,
    ) -> float:
        """
        Compute auc roc (Receiver operating characteristic) score using sklearn (one vs rest)
        :param pred: prediction array per sample. Each element shape [num_classes]
        :param target: target per sample. Each element is an integer in range [0 - num_classes)
        :param sample_weight: Optional - weight per sample for a weighted auc. Each element is  float in range [0-1]
        :param pos_class_index: the class to compute the metrics in one vs rest manner - set to 1 in binary classification
        :param max_fpr: float > 0 and <= 1, default=None
                        If not ``None``, the standardized partial AUC over the range [0, max_fpr] is returned.
        :return auc Receiver operating characteristic score
        """
        if not isinstance(pred[0], np.ndarray):
            pred = [np.array(p) for p in pred]
            pos_class_index = 1
            y_score = np.asarray(pred)
        else:
            if pos_class_index < 0:
                pos_class_index = pred[0].shape[0] - 1
            y_score = np.asarray(pred)[:, pos_class_index]

        return metrics.roc_auc_score(
            y_score=y_score, y_true=np.asarray(target) == pos_class_index, sample_weight=sample_weight, max_fpr=max_fpr
        )

    @staticmethod
    def roc_curve(
        pred: Sequence[Union[np.ndarray, float]],
        target: Sequence[Union[np.ndarray, int]],
        class_names: Sequence[str],
        sample_weight: Optional[Sequence[Union[np.ndarray, float]]] = None,
        output_filename: Optional[str] = None,
    ) -> Dict:
        """
        Multi class version for roc curve
        :param pred: List of arrays of shape [NUM_CLASSES]
        :param target: List of arrays specifying the target class per sample
        :return: saving roc curve to a file and return the input to figure to a dictionary
        """
        # if class_names not specified assume binary classification
        if class_names is None:
            class_names = [None, "Positive"]
        # extract info for the plot
        results = {}
        for cls, cls_name in enumerate(class_names):
            if cls_name is None:
                continue
            fpr, tpr, _ = sklearn.metrics.roc_curve(
                target, np.array(pred)[:, cls], sample_weight=sample_weight, pos_label=cls
            )
            auc = sklearn.metrics.auc(fpr, tpr)
            results[cls_name] = {"fpr": fpr, "tpr": tpr, "auc": auc}

        # display
        if output_filename is not None:
            for cls_name, cls_res in results.items():
                plt.plot(cls_res["fpr"], cls_res["tpr"], label=f'{cls_name}(auc={cls_res["auc"]:0.2f})')

            plt.title("ROC curve")
            plt.xlabel("FPR")
            plt.ylabel("TPR")
            plt.legend()
            plt.savefig(output_filename)
            plt.close()

        return results

    @staticmethod
    def auc_pr(
        pred: Sequence[Union[np.ndarray, float]],
        target: Sequence[Union[np.ndarray, int]],
        sample_weight: Optional[Sequence[Union[np.ndarray, float]]] = None,
        pos_class_index: int = -1,
    ) -> float:
        """
        Compute auc pr (precision-recall) score using sklearn (one vs rest)
        :param pred: prediction array per sample. Each element shape [num_classes]
        :param target: target per sample. Each element is an integer in range [0 - num_classes)
        :param sample_weight: Optional - weight per sample for a weighted auc. Each element is  float in range [0-1]
        :param pos_class_index: the class to compute the metrics in one vs rest manner - set to 1 in binary classification
        :return auc precision recall score
        """
        if not isinstance(pred[0], np.ndarray):
            pred = [np.array(p) for p in pred]
            pos_class_index = 1
            y_score = np.asarray(pred)
        else:
            if pos_class_index < 0:
                pos_class_index = pred[0].shape[0] - 1
            y_score = np.asarray(pred)[:, pos_class_index]

        precision, recall, _ = metrics.precision_recall_curve(
            probas_pred=y_score, y_true=np.asarray(target) == pos_class_index, sample_weight=sample_weight
        )
        return metrics.auc(recall, precision)

    @staticmethod
    def accuracy(
        pred: Sequence[Union[np.ndarray, int]],
        target: Sequence[Union[np.ndarray, int]],
        sample_weight: Optional[Sequence[Union[np.ndarray, float]]] = None,
    ):
        """
        Compute accuracy score
        :param pred: class prediction. Each element is an integer in range [0 - num_classes)
        :param target: the target class. Each element is an integer in range [0 - num_classes)
        :param sample_weight: Optional - weight per sample for a weighted score. Each element is  float in range [0-1]
        :return: accuracy score
        """

        pred = np.array(pred)
        target = np.array(target)
        return metrics.accuracy_score(target, pred, sample_weight=sample_weight)

    @staticmethod
    def confusion_metrics(
        pred: Sequence[Union[np.ndarray, int]],
        target: Sequence[Union[np.ndarray, int]],
        pos_class_index: int = 1,
        metrics: Sequence[str] = tuple(),
        sample_weight: Optional[Sequence[Union[np.ndarray, float]]] = None,
    ) -> Dict[str, float]:
        """
        Compute metrics derived from one-vs-rest confusion matrix such as 'sensitivity', 'recall', 'tpr', 'specificity',  'selectivity', 'npr', 'precision', 'ppv', 'f1'
        Assuming that there are positive cases and negative cases in targets
        :param pred: class prediction. Each element is an integer in range [0 - num_classes)
        :param target: the target class. Each element is an integer in range [0 - num_classes)
        :param pos_class_index: the class to compute the metrics in one vs rest manner - set to 1 in binary classification
        :param metrics: required metrics names, options: 'sensitivity', 'recall', 'tpr', 'specificity',  'selectivity', 'npr', 'precision', 'ppv', 'f1'
        :param sample_weight: Optional - weight per sample for a weighted score. Each element is  float in range [0-1]
        :return: dictionary, including the computed values for the required metrics.
                 format: {"tp": <>, "tn": <>, "fp": <>, "fn": <>, <required metric name>: <>}
        """
        pred = np.array(pred)
        target = np.array(target)

        class_target_t = np.where(target == pos_class_index, 1, 0)
        class_pred_t = np.where(pred == pos_class_index, 1, 0)
        if sample_weight is None:
            sample_weight = np.ones_like(class_target_t)

        res = {}
        tp = (np.logical_and(class_target_t, class_pred_t) * sample_weight).sum()
        fn = (np.logical_and(class_target_t, np.logical_not(class_pred_t)) * sample_weight).sum()
        fp = (np.logical_and(np.logical_not(class_target_t), class_pred_t) * sample_weight).sum()
        tn = (np.logical_and(np.logical_not(class_target_t), np.logical_not(class_pred_t)) * sample_weight).sum()

        for metric in metrics:
            if metric in ["sensitivity", "recall", "tpr"]:
                res[metric] = tp / (tp + fn)
            elif metric in ["specificity", "selectivity", "tnr"]:
                res[metric] = tp / (tn + fp)
            elif metric in ["precision", "ppv"]:
                if tp + fp != 0:
                    res[metric] = tp / (tp + fp)
                else:
                    res[metric] = 0
            elif metric in ["f1"]:
                res[metric] = 2 * tp / (2 * tp + fp + fn)
            elif metric in ["matrix"]:
                res["tp"] = tp
                res["fn"] = fn
                res["fp"] = fp
                res["tn"] = tn
            else:
                raise Exception(f"unknown metric {metric}")

        return res

    @staticmethod
    def confusion_matrix(
        cls_pred: Sequence[int],
        target: Sequence[int],
        class_names: Sequence[str],
        sample_weight: Optional[Sequence[float]] = None,
    ) -> Dict[str, pd.DataFrame]:
        """
        Calculates Confusion Matrix (multi class version)
        :param cls_pred: sequence of class prediction
        :param target: sequence of labels
        :param class_names: string name per class
        :param sample_weight: optional, weight per sample.
        :return: {"count": <confusion matrix>, "percent" : <confusion matrix - percent>)
        Confusion matrix whose i-th row and j-th column entry indicates the number of samples with true label being i-th class and predicted label being j-th class.
        """
        conf_matrix = sklearn.metrics.confusion_matrix(y_true=target, y_pred=cls_pred, sample_weight=sample_weight)
        conf_matrix_count = pd.DataFrame(conf_matrix, columns=class_names, index=class_names)
        conf_matrix_total = conf_matrix.sum(axis=1)
        conf_matrix_count["total"] = conf_matrix_total
        conf_matrix_percent = pd.DataFrame(
            conf_matrix / conf_matrix_total[:, None], columns=class_names, index=class_names
        )

        return {"count": conf_matrix_count, "percent": conf_matrix_percent}

    @staticmethod
    def multi_class_bs(pred: np.ndarray, target: np.ndarray) -> float:
        """
        Brier Score:
        bs = 1/N * SUM_n SUM_c (pred_{n,c} - target_{n,c})^2
        :param pred: probability score. Expected Shape [N, C]
        :param target: target class (int) per sample. Expected Shape [N]
        """
        # create one hot vector
        target_one_hot = np.zeros_like(pred)
        target_one_hot[np.arange(target_one_hot.shape[0]), target] = 1

        return float(np.mean(np.sum((pred - target_one_hot) ** 2, axis=1)))

    @staticmethod
    def multi_class_bss(pred: Sequence[np.ndarray], target: Sequence[np.ndarray]) -> float:
        """
        Brier Skill Score:
        bss = 1 - bs / bs_{ref}

        bs_{ref} will be computed for a model that makes a predictions according to the prevalance of each class in dataset

        :param pred: probability score. Expected Shape [N, C]
        :param target: target class (int) per sample. Expected Shape [N]
        """
        if isinstance(pred[0], np.ndarray) and pred[0].shape[0] > 1:
            pred = np.array(pred)
        else:
            # binary case
            pred = np.array(pred)
            pred = np.stack((1 - pred, pred), axis=-1)
        target = np.array(target)

        # BS
        bs = MetricsLibClass.multi_class_bs(pred, target)

        # no skill BS
        no_skill_prediction = [(target == target_cls).sum() / target.shape[0] for target_cls in range(pred.shape[-1])]
        no_skill_predictions = np.tile(np.array(no_skill_prediction), (pred.shape[0], 1))
        bs_ref = MetricsLibClass.multi_class_bs(no_skill_predictions, target)

        return 1.0 - bs / bs_ref

    @staticmethod
    def convert_probabilities_to_class(
        pred: Sequence[Union[np.ndarray, float]], operation_point: Union[float, Sequence[Tuple[int, float]]]
    ) -> np.array:
        """
        convert probabilities to class prediction
        :param pred: sequence of numpy arrays / floats of shape [NUM_CLASSES]
        :param operation_point: list of tuples (class_idx, threshold) or empty sequence for argmax
        :return: array of class predictions
        """
        if isinstance(pred[0], np.ndarray) and pred[0].shape[0] > 1:
            pred = np.array(pred)
        else:
            # binary case
            pred = np.array(pred)
            pred = np.stack((1 - pred, pred), axis=-1)

        # if no threshold specified, simply apply argmax
        if operation_point is None or (isinstance(operation_point, Sequence) and len(operation_point) == 0):
            return np.argmax(pred, -1)

        # binary operation point
        if isinstance(operation_point, float):
            if pred[0].shape[0] == 2:
                return np.where(pred[:, 1] > operation_point, 1, 0)
            elif pred[0].shape[0] == 1:
                return np.where(pred > operation_point, 1, 0)
            else:
                raise Exception(f"Error - got single float as an operation point for multiclass prediction")

        # convert according to thresholds
        output_class = np.array([-1 for x in range(len(pred))])
        for thr in operation_point:
            class_idx = thr[0]
            class_thr = thr[1]
            # argmax
            if class_idx == "argmax":
                output_class[output_class == -1] = np.argmax(pred, -1)[output_class == -1]

            # among all the samples which not already predicted, set the ones that cross the threshold with this class
            target_idx = np.argwhere(np.logical_and(pred[:, class_idx] > class_thr, output_class == -1))
            output_class[target_idx] = class_idx

        return output_class



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import os
import pathlib
from tempfile import mkdtemp
from typing import Any, Dict

from collections import OrderedDict

import pandas as pd
import numpy as np

from fuse.utils import set_seed

from fuse.eval.metrics.metrics_common import GroupAnalysis, CI, Filter
from fuse.eval.metrics.metrics_model_comparison import PairedBootstrap
from fuse.eval.metrics.classification.metrics_classification_common import (
    MetricAUCPR,
    MetricAUCROC,
    MetricAccuracy,
    MetricConfusion,
    MetricConfusionMatrix,
    MetricBSS,
    MetricROCCurve,
)
from fuse.eval.metrics.classification.metrics_model_comparison_common import (
    MetricDelongsTest,
    MetricMcnemarsTest,
)
from fuse.eval.metrics.classification.metrics_thresholding_common import MetricApplyThresholds
from fuse.eval.metrics.classification.metrics_calibration_common import (
    MetricReliabilityDiagram,
    MetricECE,
    MetricFindTemperature,
    MetricApplyTemperature,
)
from fuse.eval.evaluator import EvaluatorDefault


def example_0() -> Dict[str, Any]:
    """
    Simple evaluation example for binary classification task
    Input is a single dataframe pickle file including 3 columns: "id", "pred" (numpy arrays) and "target"
    """
    # path to prediction and target files
    dir_path = pathlib.Path(__file__).parent.resolve()
    input_filename = os.path.join(dir_path, "inputs/example0.pickle")

    # list of metrics
    metrics = OrderedDict(
        [
            ("auc", MetricAUCROC(pred="pred", target="target")),
        ]
    )

    # read files
    data = input_filename
    evaluator = EvaluatorDefault()
    results = evaluator.eval(
        ids=None, data=data, metrics=metrics
    )  # ids == None -> run evaluation on all available samples

    return results


def example_1() -> Dict[str, Any]:
    """
    Simple evaluation example
    Inputs are two .csv files: one including predictions and one targets
    The predictions requires a simple pre-processing to create prediction in a format of vector of probabilities
    """
    # path to prediction and target files
    dir_path = pathlib.Path(__file__).parent.resolve()
    prediction_filename = os.path.join(dir_path, "inputs/example1_predictions.csv")
    targets_filename = os.path.join(dir_path, "inputs/example1_targets.csv")

    # define data
    data = {"pred": prediction_filename, "target": targets_filename}

    # pre collect function to change the format
    def pre_collect_process(sample_dict: dict) -> dict:
        # convert scores to numpy array
        task1_pred = []
        for cls_name in ("NoAT", "CanAT"):
            task1_pred.append(sample_dict[f"pred.{cls_name}-score"])
        task1_pred_array = np.array(task1_pred)
        sample_dict["pred.array"] = task1_pred_array

        return sample_dict

    # list of metrics
    metrics = OrderedDict(
        [
            (
                "auc",
                MetricAUCROC(
                    pred="pred.array", target="target.Task1-target", pre_collect_process_func=pre_collect_process
                ),
            ),
        ]
    )

    # specify just the list you want to evaluate - predictions may contain more samples that will be ignored
    task_target_df = pd.read_csv(targets_filename)
    ids = list(task_target_df["id"])

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=ids, data=data, metrics=metrics)

    return results


def example_2():
    """
    Cross validation example - evaluation the entire data, built from few folds at once
    Multiple inference files - each include prediction of a different fold - binary predictions (single probability)
    Single target file with labels for all the folds
    """
    # path to prediction and target files
    dir_path = pathlib.Path(__file__).parent.resolve()
    prediction_fold0_filename = os.path.join(dir_path, "inputs/example2_predictions_fold0.csv")
    prediction_fold1_filename = os.path.join(dir_path, "inputs/example2_predictions_fold1.csv")
    targets_filename = os.path.join(dir_path, "inputs/example2_targets.csv")

    # define data
    data = {"pred": [prediction_fold0_filename, prediction_fold1_filename], "target": targets_filename}

    # list of metrics
    metrics = OrderedDict(
        [
            ("auc", MetricAUCROC(pred="pred.CanAT-score", target="target.Task1-target")),
            (
                "auc_per_fold",
                GroupAnalysis(
                    MetricAUCROC(pred="pred.CanAT-score", target="target.Task1-target"), group="pred.evaluator_fold"
                ),
            ),
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data, metrics=metrics)

    return results


def example_3():
    """
    General group analysis example - compute the AUC for each group separately.
    In this case the grouping is done according to gender
    """
    data = {
        "pred": [0.1, 0.2, 0.6, 0.7, 0.8, 0.3, 0.6, 0.2, 0.7, 0.9],
        "target": [0, 0, 1, 1, 1, 0, 0, 1, 1, 1],
        "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        "gender": ["female", "female", "female", "female", "female", "male", "male", "male", "male", "male"],
    }
    data = pd.DataFrame(data)

    metrics = OrderedDict(
        [("auc_per_group", GroupAnalysis(MetricAUCROC(pred="pred", target="target"), group="gender"))]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data, metrics=metrics)

    return results


def example_4() -> Dict[str, Any]:
    """
    Simple evaluation example with Confidence Interval
    Inputs are two .csv files: one including predictions and one targets
    """
    # set seed
    seed = 1234
    # path to prediction and target files - reuse example1 inputs
    dir_path = pathlib.Path(__file__).parent.resolve()
    prediction_filename = os.path.join(dir_path, "inputs/example1_predictions.csv")
    targets_filename = os.path.join(dir_path, "inputs/example1_targets.csv")

    # define data
    data = {"pred": prediction_filename, "target": targets_filename}

    # list of metrics
    metrics = OrderedDict(
        [
            (
                "auc",
                CI(
                    MetricAUCROC(pred="pred.CanAT-score", target="target.Task1-target"),
                    stratum="target.Task1-target",
                    rnd_seed=seed,
                ),
            ),
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data, metrics=metrics)

    return results


def example_5():
    """
    Model comparison using paired bootstrap metric
    Compare model a binary classification sensitivity to model b binary classification sensitivity
    """
    # set seed
    seed = 0

    # define data
    data = {
        "id": [0, 1, 2, 3, 4],
        "model_a_pred": [0.4, 0.3, 0.7, 0.8, 0.0],
        "model_b_pred": [0.4, 0.3, 0.7, 0.2, 1.0],
        "target": [0, 1, 1, 1, 0],
    }

    data_df = pd.DataFrame(data)

    # list of metrics
    metric_model_test = MetricConfusion(pred="results:metrics.apply_thresh_a.cls_pred", target="target")
    metric_model_reference = MetricConfusion(pred="results:metrics.apply_thresh_b.cls_pred", target="target")

    metrics = OrderedDict(
        [
            ("apply_thresh_a", MetricApplyThresholds(pred="model_a_pred", operation_point=0.5)),
            ("apply_thresh_b", MetricApplyThresholds(pred="model_b_pred", operation_point=0.5)),
            (
                "compare_a_to_b",
                PairedBootstrap(
                    metric_model_test,
                    metric_model_reference,
                    stratum="target",
                    metric_keys_to_compare=["sensitivity"],
                    rnd_seed=seed,
                ),
            ),
        ]
    )

    # read files
    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data_df, metrics=metrics)

    return results


def example_6() -> Dict:
    """
    Simple test of the DeLong's test implementation
    Also "naively" test the multiclass mode (one vs. all) by simply extending the
    binary classifier prediction vectors into a 2D matrix
    For this simple example, DeLong's test was computed manually in this blog post by Rachel Draelos:
    https://glassboxmedicine.com/2020/02/04/comparing-aucs-of-machine-learning-models-with-delongs-test/
    """
    target = np.array([0, 0, 1, 1, 1])
    pred1 = np.array([0.1, 0.2, 0.6, 0.7, 0.8])
    pred2 = np.array([0.3, 0.6, 0.2, 0.7, 0.9])

    # convert to Nx2 for multi-class generalization:
    pred1 = np.stack((1 - pred1, pred1), 1)
    pred2 = np.stack((1 - pred2, pred2), 1)
    index = np.arange(0, len(target))

    # convert to dataframes:
    pred_df = pd.DataFrame(columns=["pred1", "pred2", "id"])
    pred_df["pred1"] = list(pred1)
    pred_df["pred2"] = list(pred2)
    pred_df["id"] = index
    target_df = pd.DataFrame(columns=["target", "id"])
    target_df["target"] = target
    target_df["id"] = index
    data = {"pred": pred_df, "target": target_df}

    # list of metrics
    metrics = OrderedDict(
        [
            (
                "delongs_test",
                MetricDelongsTest(
                    target="target.target", class_names=["negative", "positive"], pred1="pred.pred1", pred2="pred.pred2"
                ),
            ),
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data, metrics=metrics)

    return results


def example_7() -> Dict:
    """
    Another example for testing the DeLong's test implementation. This time in "binary classifier" mode
    The sample data in this example was used in the above blog post and verified against an R implementation.
    Three different sources: dataframe per model prediction and a target dataframe
    """
    target = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1])
    pred1 = np.array([0.1, 0.2, 0.05, 0.3, 0.1, 0.6, 0.6, 0.7, 0.8, 0.99, 0.8, 0.67, 0.5])
    pred2 = np.array([0.3, 0.6, 0.2, 0.1, 0.1, 0.9, 0.23, 0.7, 0.9, 0.4, 0.77, 0.3, 0.89])

    ids = np.arange(0, len(target))

    # convert to dataframes:
    pred1_df = pd.DataFrame(columns=["pred1", "pred2", "id"])
    pred1_df["output"] = pred1
    pred1_df["id"] = ids

    pred2_df = pd.DataFrame(columns=["pred1", "pred2", "id"])
    pred2_df["output"] = pred2
    pred2_df["id"] = ids

    target_df = pd.DataFrame(columns=["target", "id"])
    target_df["target"] = target
    target_df["id"] = ids
    data = {"pred1": pred1_df, "pred2": pred2_df, "target": target_df}

    # list of metrics
    metrics = OrderedDict(
        [
            ("delongs_test", MetricDelongsTest(target="target.target", pred1="pred1.output", pred2="pred2.output")),
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data, metrics=metrics)

    return results


def example_8():
    """
    Classification Multiclass example: five classes evaluation with metrics AUC-ROC AUC-PR, sensitivity, specificity and precision
    Input: one .csv prediction file that requires processing to convert the predictions to numpy array and one target file.
    """
    # path to prediction and target files
    dir_path = pathlib.Path(__file__).parent.resolve()
    prediction_filename = os.path.join(dir_path, "inputs/example7_predictions.csv")
    targets_filename = os.path.join(dir_path, "inputs/example1_targets.csv")  # same target file as in example1
    data = {"target": targets_filename, "pred": prediction_filename}

    class_names = ["B", "LR", "IR", "HR", "VHR"]

    # pre collect function to change the format
    def pre_collect_process(sample_dict: dict) -> dict:
        # convert scores to numpy array
        task2_pred = []
        for cls_name in class_names:
            task2_pred.append(sample_dict[f"pred.{cls_name}-score"])
        task2_pred_array = np.array(task2_pred)
        sample_dict["pred.output"] = task2_pred_array
        sample_dict["pred.output_class"] = task2_pred_array.argmax()

        return sample_dict

    # list of metrics
    metrics = OrderedDict(
        [
            (
                "auc",
                MetricAUCROC(
                    pred="pred.output",
                    target="target.Task2-target",
                    class_names=class_names,
                    pre_collect_process_func=pre_collect_process,
                ),
            ),
            (
                "auc_pr",
                MetricAUCPR(
                    pred="pred.output",
                    target="target.Task2-target",
                    class_names=class_names,
                    pre_collect_process_func=pre_collect_process,
                ),
            ),
            (
                "confusion",
                MetricConfusion(
                    pred="pred.output_class",
                    target="target.Task2-target",
                    class_names=class_names,
                    pre_collect_process_func=pre_collect_process,
                    metrics=["sensitivity", "specificity", "precision"],
                ),
            ),  # default operation point is argmax
            (
                "accuracy",
                MetricAccuracy(
                    pred="pred.output_class", target="target.Task2-target", pre_collect_process_func=pre_collect_process
                ),
            ),  # default operation point is argmax
            (
                "confusion_matrix",
                MetricConfusionMatrix(
                    cls_pred="pred.output_class",
                    target="target.Task2-target",
                    class_names=class_names,
                    pre_collect_process_func=pre_collect_process,
                ),
            ),
            (
                "bss",
                MetricBSS(
                    pred="pred.output", target="target.Task2-target", pre_collect_process_func=pre_collect_process
                ),
            ),
            (
                "roc",
                MetricROCCurve(
                    pred="pred.output",
                    target="target.Task2-target",
                    class_names=class_names,
                    output_filename="roc.png",
                    pre_collect_process_func=pre_collect_process,
                ),
            ),
        ]
    )

    # read files
    evaluator = EvaluatorDefault()
    results = evaluator.eval(
        ids=None, data=data, metrics=metrics
    )  # ids == None -> run evaluation on all available samples

    return results


def example_9():
    """
    Classification example with single-process iterator.
    This example requires fuse-med-ml-data and torchvision packages installed
    """
    import torchvision

    # set seed
    set_seed(1234)

    # Create dataset
    # mnist download path - MNIST_DATA_PATH defined in cicd pipeline
    mnist_data_path = os.environ.get("MNIST_DATA_PATH", mkdtemp(prefix="mnist"))
    torch_dataset = torchvision.datasets.MNIST(mnist_data_path, download=True, train=False)

    # define iterator
    def data_iter():
        for sample_index, (image, label) in enumerate(torch_dataset):
            sample_dict = {}
            sample_dict["id"] = sample_index
            sample_dict["pred"] = np.random.randint(0, 10)
            sample_dict["label"] = label
            yield sample_dict

        # list of metrics

    metrics = OrderedDict(
        [
            (
                "accuracy",
                MetricAccuracy(pred="pred", target="label"),
            ),  # operation_point=None -> no need to convert pred from probabilities to class
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(
        ids=None, data=data_iter(), batch_size=5, metrics=metrics
    )  # ids == None -> run evaluation on all available samples

    return results


def example_10() -> Dict:
    """
    Test of McNemar's test implementation
    """

    # Toy example:

    # The correct target (class) labels (optional)
    ground_truth = [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
    # in case ground_truth is not provided (variable set to None),
    # the significance test checks the similarity between model predictions only.

    # Class labels predicted by model 1
    cls_pred1 = [0, 1, 1, 1, 0, 1, 1, 0, 1, 1]
    # Class labels predicted by model 2
    cls_pred2 = [1, 1, 0, 1, 0, 1, 1, 0, 0, 1]

    index = np.arange(0, len(cls_pred1))
    # convert to dataframes:
    df = pd.DataFrame(columns=["cls_pred1", "cls_pred2", "id"])
    df["cls_pred1"] = cls_pred1
    df["cls_pred2"] = cls_pred2
    df["ground_truth"] = ground_truth
    df["id"] = index
    data = {"data": df}

    # list of metrics
    metrics = OrderedDict(
        [
            (
                "mcnemars_test",
                MetricMcnemarsTest(pred1="data.cls_pred1", pred2="data.cls_pred2", target="data.ground_truth"),
            ),
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data, metrics=metrics)

    return results


def example_11() -> Dict:
    """
    Sub group analysis example
    """
    # define data
    data = {
        "id": [0, 1, 2, 3, 4],
        "pred": [0, 1, 0, 0, 0],
        "gender": ["male", "female", "female", "male", "female"],
        "target": [0, 1, 1, 1, 0],
    }
    data_df = pd.DataFrame(data)

    def pre_collect_process(sample_dict: dict) -> dict:
        sample_dict["filter"] = sample_dict["gender"] != "male"
        return sample_dict

    acc = MetricAccuracy(pred="pred", target="target")
    metrics = OrderedDict(
        [
            (
                "accuracy",
                Filter(acc, "filter", pre_collect_process_func=pre_collect_process),
            ),  # operation_point=None -> no need to convert pred from probabilities to class
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(
        ids=None, data=data_df, metrics=metrics
    )  # ids == None -> run evaluation on all available samples

    return results


def example_12() -> Dict:
    """
    Example of a metric pipeline which includes a per-sample metric/operation.
    First, we apply a simple thresholding operation (per sample "metric"/operation) to generate class predictions.
    Then, we apply the accuracy metric on the class predictions vs. targets.
    """
    target = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1])
    pred = np.array([0.3, 0.6, 0.2, 0.1, 0.1, 0.9, 0.23, 0.7, 0.9, 0.4, 0.77, 0.3, 0.89])

    ids = np.arange(0, len(target))

    # convert to dataframes:
    pred_df = pd.DataFrame(columns=["pred", "id"])
    pred_df["pred"] = pred
    pred_df["id"] = ids

    target_df = pd.DataFrame(columns=["target", "id"])
    target_df["target"] = target
    target_df["id"] = ids
    data = {"pred": pred_df, "target": target_df}

    metrics = OrderedDict(
        [
            ("apply_thresh", MetricApplyThresholds(pred="pred.pred", operation_point=0.5)),
            ("acc", MetricAccuracy(pred="results:metrics.apply_thresh.cls_pred", target="target.target")),
        ]
    )
    # read files
    evaluator = EvaluatorDefault()
    results = evaluator.eval(
        ids=None, data=data, metrics=metrics
    )  # ids == None -> run evaluation on all available samples

    return results


def example_13() -> Dict:
    """
    Test reliability diagram and ECE metrics
    We use multi-class input data as in example_7:
    One .csv prediction file that requires processing to convert the predictions to numpy array and one target file.

    We define a metric pipeline. First the reliability diagram is computed on the original predictions.
    Then, we compute the expected calibration error (ECE). Then, we find an optimal "temperature" value to calibrate the logits.
    Then, we apply this temperature scale on the original logits, to obtain calibrated ones.
    Finally, we compute the ECE and reliability diagram for the calibrated predictions.
    """
    # path to prediction and target files
    dir_path = pathlib.Path(__file__).parent.resolve()
    prediction_filename = os.path.join(dir_path, "inputs/example7_predictions.csv")
    targets_filename = os.path.join(dir_path, "inputs/example1_targets.csv")  # same target file as in example1
    data = {"target": targets_filename, "pred": prediction_filename}

    class_names = ["B", "LR", "IR", "HR", "VHR"]
    num_bins = 10
    num_quantiles = None

    # pre collect function to change the format
    def pre_collect_process(sample_dict: dict) -> dict:
        # convert scores to numpy array
        task2_pred = []
        for cls_name in class_names:
            task2_pred.append(sample_dict[f"pred.{cls_name}-score"])
        task2_pred_array = np.array(task2_pred)
        logit_array = np.log(task2_pred_array)  # "make up" logits up to a constant
        sample_dict["pred.output"] = task2_pred_array
        sample_dict["pred.logits"] = logit_array

        return sample_dict

    # list of metrics
    metrics = OrderedDict(
        [
            (
                "reliability",
                MetricReliabilityDiagram(
                    pred="pred.output",
                    target="target.Task2-target",
                    num_bins=num_bins,
                    num_quantiles=num_quantiles,
                    output_filename="reliability.png",
                    pre_collect_process_func=pre_collect_process,
                ),
            ),
            (
                "ece",
                MetricECE(
                    pred="pred.output",
                    target="target.Task2-target",
                    num_bins=num_bins,
                    num_quantiles=num_quantiles,
                    pre_collect_process_func=pre_collect_process,
                ),
            ),
            (
                "find_temperature",
                MetricFindTemperature(
                    pred="pred.logits", target="target.Task2-target", pre_collect_process_func=pre_collect_process
                ),
            ),
            (
                "apply_temperature",
                MetricApplyTemperature(
                    pred="pred.logits",
                    temperature="results:metrics.find_temperature",
                    pre_collect_process_func=pre_collect_process,
                ),
            ),
            (
                "ece_calibrated",
                MetricECE(
                    pred="results:metrics.apply_temperature",
                    target="target.Task2-target",
                    num_bins=num_bins,
                    num_quantiles=num_quantiles,
                    pre_collect_process_func=pre_collect_process,
                ),
            ),
            (
                "reliability_calibrated",
                MetricReliabilityDiagram(
                    pred="results:metrics.apply_temperature",
                    target="target.Task2-target",
                    num_bins=num_bins,
                    num_quantiles=num_quantiles,
                    output_filename="reliability_calibrated.png",
                    pre_collect_process_func=pre_collect_process,
                ),
            ),
        ]
    )

    # read files
    evaluator = EvaluatorDefault()
    results = evaluator.eval(
        ids=None, data=data, metrics=metrics
    )  # ids == None -> run evaluation on all available samples

    return results
